"confirmation_analysis.py"

import pathlib
import yaml

from evidence_seeker.datamodels import CheckedClaim

from .config import ConfirmationAnalyzerConfig
from .workflows import SimpleConfirmationAnalysisWorkflow


class ConfirmationAnalyzer:
    def __init__(self, config: ConfirmationAnalyzerConfig | None = None, **kwargs):
        if config is None:
            config = ConfirmationAnalyzerConfig()
        self.config = config
        self.workflow = SimpleConfirmationAnalysisWorkflow(
            config=config, **kwargs
        )

    async def __call__(self, claim: CheckedClaim) -> CheckedClaim:
        coros = [
            (
                document.uid,
                await self.workflow.run(
                    clarified_claim=claim, evidence_item=document.text
                ),
            )
            for document in claim.documents
        ]
        claim.confirmation_by_document = {
            uid: wf_result for uid, wf_result in coros
        }
        return claim

    @staticmethod
    def from_config_file(config_file: str):
        path = pathlib.Path(config_file)
        config = ConfirmationAnalyzerConfig(**yaml.safe_load(path.read_text()))
        return ConfirmationAnalyzer(config=config)



"confirmation_aggregation"

import numpy as np
import pathlib
import pydantic
import yaml

from evidence_seeker.datamodels import CheckedClaim
from evidence_seeker.datamodels import ConfirmationLevel


def confirmation_level(degree_of_confirmation: float) -> ConfirmationLevel:
    if degree_of_confirmation > 0.6:
        return ConfirmationLevel.STRONGLY_CONFIRMED
    if degree_of_confirmation > 0.4:
        return ConfirmationLevel.CONFIRMED
    if degree_of_confirmation > 0.2:
        return ConfirmationLevel.WEAKLY_CONFIRMED
    if degree_of_confirmation < -0.6:
        return ConfirmationLevel.STRONGLY_DISCONFIRMED
    if degree_of_confirmation < -0.4:
        return ConfirmationLevel.DISCONFIRMED
    if degree_of_confirmation < -0.2:
        return ConfirmationLevel.WEAKLY_DISCONFIRMED
    return ConfirmationLevel.INCONCLUSIVE_CONFIRMATION


class ConfirmationAggregationConfig(pydantic.BaseModel):
    confirmation_threshold: float = 0.2


class ConfirmationAggregator:
    def __init__(self, config: ConfirmationAggregationConfig | None = None):
        if config is None:
            config = ConfirmationAggregationConfig()
        self.config = config

    async def verbalize_confirmation(self, claim: CheckedClaim) -> str | None:
        if claim.average_confirmation is None:
            return None
        if claim.average_confirmation > 0.6:
            return "The claim is strongly confirmed."
        if claim.average_confirmation > 0.4:
            return "The claim is confirmed."
        if claim.average_confirmation > 0.2:
            return "The claim is weakly confirmed."
        if claim.average_confirmation < -0.6:
            return "The claim is strongly disconfirmed."
        if claim.average_confirmation < -0.4:
            return "The claim is disconfirmed."
        if claim.average_confirmation < -0.2:
            return "The claim is weakly disconfirmed."
        return "The claim is neither confirmed nor disconfirmed."

    async def __call__(self, claim: CheckedClaim) -> CheckedClaim:
        relevant_conf_by_docs = claim.confirmation_by_document or {}
        relevant_conf_by_docs = {
            k: c
            for k, c in relevant_conf_by_docs.items()
            if abs(c) > self.config.confirmation_threshold
        }
        claim.n_evidence = len(relevant_conf_by_docs)
        claim.average_confirmation = float(
            np.mean(list(relevant_conf_by_docs.values())) if relevant_conf_by_docs else np.nan
        )
        claim.confirmation_level = confirmation_level(
            claim.average_confirmation
        )
        claim.evidential_uncertainty = float(
            np.var(list(relevant_conf_by_docs.values())) if relevant_conf_by_docs else np.nan
        )
        claim.verbalized_confirmation = await self.verbalize_confirmation(claim)

        return claim

    @staticmethod
    def from_config_file(config_file: str):
        path = pathlib.Path(config_file)
        config = ConfirmationAggregationConfig(**yaml.safe_load(path.read_text()))
        return ConfirmationAggregator(config=config)



"preprocessing.py"

import pathlib
import yaml

from evidence_seeker.datamodels import CheckedClaim
from evidence_seeker.preprocessing.workflows import PreprocessingWorkflow
from evidence_seeker.preprocessing.config import ClaimPreprocessingConfig


class ClaimPreprocessor:

    def __init__(
        self, config: ClaimPreprocessingConfig,
        **kwargs
    ):

        self.config = config
        self.workflow = PreprocessingWorkflow(
            config=config, **kwargs
        )

    async def __call__(self, claim: str) -> list[CheckedClaim]:
        workflow_result = await self.workflow.run(claim=claim)
        return workflow_result

    @staticmethod
    def from_config_file(config_file: str):
        path = pathlib.Path(config_file)
        config = ClaimPreprocessingConfig(**yaml.safe_load(path.read_text()))
        return ClaimPreprocessor(config=config)




import pydantic
from typing import Any, Dict


class EvidenceSeekerClientConfig(pydantic.BaseModel):
    config_version: str = "v0.1"
    description: str = "First version of a configuration file for the Evidence Seeker client."

    init_directory_structure: Dict[str, Any] = {
        "config": {},
        "knowledge_base": {},
        "scripts": {},
        "index": {},
        "logs": {},
    }
    config_directory: str = "config"
    index_directory: str = "logs"



"confirmation_analysis.py"

from typing import Any, Dict, List, Optional, Self
from loguru import logger
from llama_index.core import ChatPromptTemplate

import re
import pydantic
import enum

from evidence_seeker.backend import GuidanceType

# TODO: validation for `used_model_key`: Check whether corresponding model is defined


class LogProbsType(enum.Enum):
    OPENAI_LIKE = "openai_like"
    ESTIMATE = "estimate"


def json_schema(pattern: str) -> Dict[str, Any]:
    return {
        "properties": {
            "answer": {
                "type": "string",
                "pattern": pattern
            }
        },
        "required": [
            "answer",
        ]
    }

class ConfirmationAnalyzerModelStepConfig(pydantic.BaseModel):
    prompt_template: str
    system_prompt: str | None = None


class MultipleChoiceTaskStepConfig(ConfirmationAnalyzerModelStepConfig):
    """Configuration for the multiple choice task step."""
    prompt_template: str = (
        "Your task is to sum up the results of a rich textual entailment analysis.\n"
        "\n"
        "<TEXT>{evidence_item}</TEXT>\n"
        "\n"
        "<HYPOTHESIS>{statement}</HYPOTHESIS>\n"
        "\n"
        "Our previous analysis has yielded the following result:\n"
        "\n"
        "<RESULT>\n"
        "{freetext_confirmation_analysis}\n"
        "</RESULT>\n"
        "\n"
        "Please sum up this result by deciding which of the following choices is correct. "
        "Just answer with the label of the correct choice.\n"
        "\n"
        "{answer_options}\n"
        "\n"
    )
    # TODO: As set of strings
    answer_options: Optional[List[str]] = [
        "Entailment: The TEXT provides sufficient evidence to support the HYPOTHESIS.",
        "Contradiction: The TEXT provides evidence that contradicts the HYPOTHESIS.",
        "Neutral: The TEXT neither supports nor contradicts the HYPOTHESIS.",
    ]
    answer_labels: Optional[List[str]] = ["A", "B", "C"]
    claim_option: Optional[str] = (
        "Entailment: The TEXT provides sufficient evidence to support the HYPOTHESIS."
    )
    n_repetitions_mcq: int = 1
    # Used as delimiter between an answer label and an answer option
    delim_str: Optional[str] = ""
    # Fields used for constrained decoding
    guidance_type: Optional[str] = GuidanceType.JSON.value
    # TODO: refactor as computed field based on answer labels
    constrained_decoding_regex: Optional[str] = None
    # TODO: refactor as computed field based on answer labels
    constrained_decoding_grammar: Optional[str] = None
    # used for regex-based validation of the output for `GuidanceType.PROMPTED`
    # TODO: refactor as computed field based on answer labels
    validation_regex: Optional[str] = None
    # log probs
    logprobs_type: Optional[str] = LogProbsType.OPENAI_LIKE.value

    # JSON schema for JSON Guidance
    @pydantic.computed_field
    @property
    def json_schema(self) -> Optional[str | Dict[str, Any]]:
        if (
            self.guidance_type == GuidanceType.JSON.value
            and (self.answer_labels is not None)
            and len(self.answer_labels) != 0
        ):
            return json_schema(pattern=rf"^({'|'.join(self.answer_labels)})$")
        return None

    # Validation of answer labels for JSON Guidance
    @pydantic.model_validator(mode='after')
    def check_answer_labels(self) -> Self:
        if self.guidance_type == GuidanceType.JSON.value:
            if self.answer_labels is None or len(self.answer_labels) == 0:
                raise ValueError(
                    'Please provide possible answer labels for multiple '
                    'choice tasks when using JSON Guidance.'
                )
            else:
                seen = set()
                seen_twice = set(
                    x for x in self.answer_labels if x in seen or seen.add(x)
                )
                valid_labels = True; [valid_labels := valid_labels & (re.match(r"^[a-zA-Z0-9]$", l) is not None) for l in self.answer_labels]
                if len(seen_twice) != 0 or not valid_labels:
                    raise ValueError(
                        "JSON Guidance assumes unique and single characters "
                        "as answer labels. Possible characters are ASCII "
                        "characters in the ranges of a-z, A-Z and 0-9."
                    )
        return self


class ConfirmationAnalyzerStepConfig(pydantic.BaseModel):
    name: str
    description: str | None = None
    used_model_key: str | None = None
    llm_specific_configs: Dict[
        str, ConfirmationAnalyzerModelStepConfig | MultipleChoiceTaskStepConfig
    ] = dict()


class ConfirmationAnalyzerConfig(pydantic.BaseModel):
    config_version: str = "v0.2"
    description: str = (
        "Configuration of EvidenceSeeker's confirmation analyzer component."
    )
    system_prompt: str = (
        "You are a helpful assistant with outstanding expertise in critical thinking and logico-semantic analysis.\n"
        "You have a background in philosophy and experience in fact checking and debate analysis.\n"
        "You read instructions carefully and follow them precisely. You give concise and clear answers."
    )
    timeout: int = 900
    # Whether or not the workflow/pipeline should print additional informative 
    # messages during execution.
    verbose: bool = False
    used_model_key: Optional[str] = None
    env_file: str | None = None

    @pydantic.model_validator(mode='after')
    def load_env_file(
        cls,
        config: 'ConfirmationAnalyzerConfig'
    ) -> 'ConfirmationAnalyzerConfig':
        if config.env_file is None:
            logger.warning(
                "No environment file with API keys specified for confirmation "
                "analyzer. Please set 'env_file' to a valid path if you want "
                "to load environment variables from a file."
            )
        else:
            # check if the env file exists
            from os import path
            if not path.exists(config.env_file):
                err_msg = (
                    f"Environment file '{config.env_file}' does not exist. "
                    "Please provide a valid path to the environment file. "
                    "Or set it to None if you don't need it and set the "
                    "API keys in other ways as environment variables."
                )
                logger.warning(err_msg)
            else:
                # load the env file
                from dotenv import load_dotenv
                load_dotenv(config.env_file)
            logger.info(
                f"Loaded environment variables from '{config.env_file}'"
            )

        return config

    freetext_confirmation_analysis: ConfirmationAnalyzerStepConfig = pydantic.Field(
        default_factory=lambda: ConfirmationAnalyzerStepConfig(
            name="freetext_confirmation_analysis",
            description="Instruct the assistant to carry out free-text RTE analysis.",
            llm_specific_configs={
                "default": ConfirmationAnalyzerModelStepConfig(
                    prompt_template=(
                        "Determine the relationship between the following two texts:\n"
                        "<TEXT>{evidence_item}</TEXT>\n"
                        "\n"
                        "<HYPOTHESIS>{statement}</HYPOTHESIS>\n"
                        "Does the TEXT entail, contradict, or neither entail nor contradict the HYPOTHESIS?\n"
                        "Classify the relationship as one of the following:\n"
                        "Entailment: The TEXT provides sufficient evidence to support the HYPOTHESIS.\n"
                        "Contradiction: The TEXT provides evidence that contradicts the HYPOTHESIS.\n"
                        "Neutral: The TEXT neither supports nor contradicts the HYPOTHESIS.\n"
                        "Please discuss this question thoroughly before providing your final answer."
                    )
                )
            }
        )
    )
    multiple_choice_confirmation_analysis: ConfirmationAnalyzerStepConfig = pydantic.Field(
         default_factory=lambda: ConfirmationAnalyzerStepConfig(
                name="multiple_choice_confirmation_analysis",
                description="Multiple choice RTE task given CoT trace.",
                llm_specific_configs={
                    "default": MultipleChoiceTaskStepConfig(),
                }
            ),
    )
    # TODO (?): Define Pydantic class for model. Or do we leave it at this?
    # Since we want to unpack the dict as model kwargs.
    models: Dict[str, Dict[str, Any]] = pydantic.Field(
        default_factory=lambda: dict()
    )

    # ==helper functions==
    def _step_config(
        self,
        step_config: Optional[ConfirmationAnalyzerStepConfig] = None,
        step_name: Optional[str] = None
    ) -> ConfirmationAnalyzerStepConfig:
        """Internal convenience function."""
        if step_config is None and step_name is None:
            raise ValueError("Either pass a step config or a name of the pipeline step")
        if step_config is None:
            if step_name == "multiple_choice_confirmation_analysis":
                return self.multiple_choice_confirmation_analysis
            elif step_name == "freetext_confirmation_analysis":
                return self.freetext_confirmation_analysis
            else:
                raise ValueError(f"Did not found step config for {step_name}")
        else:
            return step_config

    def get_step_config(
            self,
            step_name: Optional[str] = None,
            step_config: Optional[ConfirmationAnalyzerStepConfig] = None
    ) -> ConfirmationAnalyzerModelStepConfig | MultipleChoiceTaskStepConfig:
        """Get the model specific step config for the given step name.

        The requested `ConfirmationAnalyzerModelStepConfig` is determined
        by either the provided `step_name` or the provided `step_config`.
        If both are given, the `step_config` is used.
        """
        step_config = self._step_config(step_config, step_name)
        # used model for this step
        if step_config.used_model_key:
            model_key = step_config.used_model_key
        else:
            model_key = self.used_model_key
        # do we have a model-specific config?
        if step_config.llm_specific_configs.get(model_key):
            model_specific_conf = step_config.llm_specific_configs[model_key]
        else:
            if step_config.llm_specific_configs.get("default") is None:
                msg = (
                    f"Default step config for {step_config.name} "
                    "not found in config."
                )
                logger.error(msg)
                raise ValueError(msg)
            model_specific_conf = step_config.llm_specific_configs["default"]
        return model_specific_conf

    def get_chat_template(
            self,
            step_name: Optional[str] = None,
            step_config: Optional[ConfirmationAnalyzerStepConfig] = None
    ) -> ChatPromptTemplate:
        step_config = self._step_config(step_config, step_name)
        model_specific_conf = self.get_step_config(step_config=step_config)
        prompt_template = model_specific_conf.prompt_template

        return ChatPromptTemplate.from_messages(
            [
                ("system", self.get_system_prompt(step_config=step_config)),
                ("user", prompt_template),
            ]
        )

    def get_system_prompt(
            self,
            step_name: Optional[str] = None,
            step_config: Optional[ConfirmationAnalyzerStepConfig] = None
    ) -> str:
        """Get the system prompt for a specific step of the workflow."""
        step_config = self._step_config(step_config, step_name)
        model_specific_conf = self.get_step_config(step_config=step_config)
        if model_specific_conf.system_prompt:
            return model_specific_conf.system_prompt
        else:
            return self.system_prompt

    def get_model_key(
            self,
            step_name: Optional[str] = None,
            step_config: Optional[ConfirmationAnalyzerStepConfig] = None
    ) -> str:
        """Get the model key for a specific step of the workflow."""
        step_config = self._step_config(step_config, step_name)
        if step_config.used_model_key:
            return step_config.used_model_key
        else:
            return self.used_model_key



import pydantic
from typing import Optional
import yaml
import textwrap


class AppConfig(pydantic.BaseModel):
    logging: bool = True
    # one of monthly, weekly, yearly, daily, None
    subdirectory_construction: Optional[str] = None
    confirmation_analysis_config_file: str
    preprocessing_config_file: str
    retrieval_config_file: str
    local_base: str | None = None
    result_dir: str
    repo_name: str | None = None
    write_on_github: bool = False
    github_token_name: str = "GITHUB_TOKEN"
    password_protection: bool = False
    password_env_name: str = "EVSE_APP_HASH"
    force_agreement: bool = False
    language: str = "de"
    example_inputs_file: str | None = None
    example_inputs: dict[str, list[str]] | None = {
        "de": [],
        "en": [],
    }
    markdown_template_file: str | None = None
    markdown_template: dict[str, str] | None = None
    save_markdown: bool = True

    translations: dict[str, dict[str, str]] = {
        "de": {
            "ascriptive": "zuschreibend",
            "descriptive": "deskriptiv",
            "normative": "normativ",
            "strongly_confirmed": "im hohen Maße bestätigt",
            "confirmed": "bestätigt",
            "weakly_confirmed": "im geringen Maße bestätigt",
            "strongly_disconfirmed": "im hohen Maße widerlegt",
            "disconfirmed": "widerlegt",
            "weakly_disconfirmed": "im geringen Maße widerlegt",
            "inconclusive_confirmation": "weder bestätigt noch widerlegt",
            "The claim is neither confirmed nor disconfirmed.": "Die Aussage wird weder bestätigt noch widerlegt.",
            "The claim is strongly confirmed.": "Die Aussage wird im hohen Maße bestätigt.",
            "The claim is strongly disconfirmed.": "Die Aussage wird im hohen Maße widerlegt.",
            "The claim is weakly confirmed.": "Die Aussage wird in geringem Maße bestätigt.",
            "The claim is weakly disconfirmed.": "Die Aussage wird in geringem Maße widerlegt.",
        },
        "en": {}
    }
    # Multi-language UI texts
    ui_texts: dict[str, dict[str, str]] = {
        "de": {
            "title": "🕵️‍♀️ EvidenceSeeker DemoApp",
            "info": textwrap.dedent("""
                <div style="background-color: #f8fafc; border: 1px solid #e2e8f0; border-radius: 8px; padding: 20px; margin: 10px 0;">
                    <details style="cursor: pointer;">
                        <summary style="font-weight: 600; font-size: 16px; color: #374151; margin-bottom: 15px; outline: none;">
                            📋 Informationen zur DemoApp
                        </summary>
                        <div style="margin-top: 15px; line-height: 1.6; color: #4b5563;">
                            <div style="margin-bottom: 20px;">
                                <h4 style="color: #1f2937; margin: 0 0 8px 0; font-size: 14px; font-weight: 600;">
                                    🔍 Grundidee der Pipeline:
                                </h4>
                                <p style="margin: 0 0 12px 15px; font-size: 14px;">
                                    Die Pipeline findet in einem ersten Schritt unterschiedliche Interpretationen deiner Eingabe und unterscheidet 
                                    dabei <em>deskriptive, zuschreibende und normative Aussagen</em>. Für die gefundenen deskriptiven und 
                                    zuschreibenden Interpretationen wird dann in einer <em>Wissensbasis</em> nach relevanten Textstellen gesucht 
                                    und analysiert, inwiefern die Textstellen die gefundene Interpretation bestätigen oder widerlegen. 
                                    Diese Einzelanalysen werden für jede Interpretation in Form eines <em>Bestätigungslevels</em> aggregiert. 
                                    Nähere Informationen zur Pipeline findest Du 
                                    <a href="#" style="color: #3b82f6; text-decoration: none;">hier</a>.
                                </p>
                            </div>
                            <div style="margin-bottom: 20px;">
                                <h4 style="color: #1f2937; margin: 0 0 8px 0; font-size: 14px; font-weight: 600;">
                                    🤖 Verwendete Modelle und Wissensbasis:
                                </h4>
                                <p style="margin: 0 0 12px 15px; font-size: 14px;">
                                    In dieser Demo App verwenden wir 
                                    <a href="https://huggingface.co/sentence-transformers/paraphrase-multilingual-mpnet-base-v2" style="color: #3b82f6; text-decoration: none;" target="_blank">paraphrase-multilingual-mpnet-base-v2</a> 
                                    als Embedding Modell sowie 
                                    <a href="https://huggingface.co/moonshotai/Kimi-K2-Instruct" style="color: #3b82f6; text-decoration: none;" target="_blank">Kimi-K2-Instruct</a> und 
                                    <a href="https://huggingface.co/meta-llama/Llama-3.3-70B-Instruct" style="color: #3b82f6; text-decoration: none;" target="_blank">Llama-3.3-70B-Instruct</a> als generative Sprachmodelle. Als Wissensbasis dienen 
                                    alle Ausgaben von "Aus Politik und Zeitgeschichte" aus dem Jahr 2024 
                                    (<a href="https://www.bpb.de/shop/zeitschriften/apuz/?field_filter_thema=all&field_date_content=2024&d=1" 
                                    style="color: #3b82f6; text-decoration: none;" target="_blank">Link</a>).
                                </p>
                            </div>
                            <div style="margin-bottom: 20px;">
                                <h4 style="color: #1f2937; margin: 0 0 8px 0; font-size: 14px; font-weight: 600;">
                                    💡 Beispiele:
                                </h4>
                                <p style="margin: 0 0 12px 15px; font-size: 14px;">
                                    Eingaben anderer User:innen und die entsprechenden Ergebnisse der Pipeline findest Du unter 
                                    <a href="https://debatelab.github.io/evidence-seeker-results/" 
                                    style="color: #3b82f6; text-decoration: none;" target="_blank">
                                        https://debatelab.github.io/evidence-seeker-results/
                                    </a>.
                                </p>
                            </div>
                            <div style="border-top: 1px solid #e5e7eb; padding-top: 15px; margin-top: 20px;">
                                <p style="margin: 0; font-size: 13px; color: #6b7280;">
                                    Die EvidenceSeeker Demoapp ist Teil des vom BMBFSFJ geförderten 
                                    <a href="https://compphil2mmae.github.io/research/kideku/" 
                                    style="color: #3b82f6; text-decoration: none;" target="_blank">KIdeKu Projekts</a>. 
                                    Nähere Informationen zur <em>EvidenceSeeker Boilerplate</em> findest Du 
                                    <a href="https://debatelab.github.io/evidence-seeker" 
                                    style="color: #3b82f6; text-decoration: none;" target="_blank">hier</a>.
                                </p>
                            </div>
                        </div>
                    </details>
                </div>
            """).strip(),
            "description": "**Gib eine Aussage in das Textfeld ein und lass sie durch den EvidenceSeeker prüfen:**",
            "statement_label": "Zu prüfende Aussage:",
            "random_example": "Zufälliges Beispiel",
            "check_statement": "Prüfe Aussage",
            "checking_message": "### Aussage wird geprüft... Dies könnte ein paar Minuten dauern.",
            "feedback_question": "Wie zufrieden bist du mit der Antwort?",
            "privacy_title": "Datenschutzhinweis & Disclaimer",
            "consent_info": "**Einwilligung zur Datenweiterverarbeitung (Optional)**",
            "agree_button": "Ich habe die Hinweise zur Kenntnis genommen",
            "password_label": (
                "Die EvidenceSeeker DemoApp ist passwortgeschützt. "
                "Bitte gib für den Zugriff auf die App das Passwort ein."
            ),
            "wrong_password": "Falsches Passwort. Bitte versuche es erneut.",
            "continue": "Weiter...",
            "server_error": "Etwas ist auf unserer Seite schiefgegangen :-(",
            "disclaimer_text": """
                    <div style="background-color:#fff7ed;padding:25px;border-radius: 10px;">
                    <p>⚠️ <b>Disclaimer</b></p>
                    <p>
                    <p>
                    Alle Ausgaben werden von Sprachmodellen generiert und geben nicht
                    zwangsläufig die Einschätzung oder Meinung der Entwickler:innen wieder.
                    </p>
                    <p>
                    Eingegebene Daten werden von Sprachmodellen verarbeitet. Bitte
                    beachte daher, keine personenbezogenen Daten einzugeben - auch weil
                    deine Eingaben unter Umständen gespeichert werden (siehe unten).
                    </p>
                    </p>
                    </div>
                """,
            "data_policy_text": """
                    <div style="background-color:#fff7ed;padding:25px;border-radius: 10px;">
                    <p>🗃️ <b>Datenschutzhinweis</b></p>
                    <p>
                    Auf der Seite <a href="https://debatelab.github.io/evidence-seeker-results/">https://debatelab.github.io/evidence-seeker-results/</a>
                    stellen wir beispielhaft
                    Ergebnisse dar, die von der EvidenceSeeker-Pipeline durch
                    die Interaktion mit Nutzer:innen über diese DemoApp erzeugt wurden.
                    </p>
                    <p>
                    Wir erheben und verwenden <strong>keine personenbezogenen Daten</strong>
                    (sofern sie nicht
                    über das Freitextfeld selbst eingegeben werden) und verwenden
                    <strong>nur</strong> von Nutzer:innen selbst eingegebene
                    Daten sowie den Zeitpunkt der Eingabe, die Rückgabe der Pipeline
                    und etwaiges Feedback durch die Nutzer:innen.
                    </p>
                    <p>
                    Wenn du das EvidenceSeeker Projekt damit unterstützen möchtest,
                    kannst du der Nutzung deiner Eingaben im Folgenden zustimmen.
                    </p>
                    </div>
                """,
            "consent_text": """
                    Ja, meine Anfragen an die EvidenceSeeker Pipeline und deren
                    Ergebnisse dürfen gespeichert und über das
                    EvidenceSeeker Projekt weiter verarbeitet und
                    veröffentlicht werden.
                """,

        },
        "en": {
            # TODO: Check & Revise
            "title": "🕵️‍♀️ EvidenceSeeker DemoApp",
            "description": "Enter a statement in the text field and have it checked by EvidenceSeeker:",
            "statement_label": "Statement to check:",
            "random_example": "Random Example",
            "check_statement": "Check Statement",
            "checking_message": "### Checking statement... This could take a few minutes.",
            "feedback_question": "How satisfied are you with the answer?",
            "privacy_title": "Privacy Notice & Disclaimer",
            "warning_label": "⚠️ <b>Warning</b>",
            "consent_info": "**Consent for data processing (Optional)**",
            "agree_button": "I have taken note of the information",
            "password_label": "Please enter password for access",
            "wrong_password": "Wrong password. Please try again.",
            "continue": "Continue...",
            "server_error": "Something went wrong on our end :-("
        }
    }

    @pydantic.computed_field
    @property
    def md_template(self) -> str:
        if self.markdown_template_file is None:
            if self.markdown_template is None:
                raise ValueError("No markdown template or file provided.")
            tmpl = self.markdown_template.get(self.language, None)
            if tmpl is None:
                raise ValueError(
                    "No markdown template available for the specified language."
                )
            return tmpl
        else:
            try:
                with open(self.markdown_template_file, encoding="utf-8") as f:
                    return f.read()
            except Exception:
                raise ValueError("Given 'markdown_template_file' not readable.")



    @pydantic.computed_field
    @property
    def examples(self) -> list[str]:
        if self.example_inputs_file is None:
            if self.example_inputs is None:
                raise ValueError("No example inputs or example file provided.")
            example_inputs = self.example_inputs.get(self.language, [])
            if not example_inputs:
                raise ValueError(
                    "No example inputs available for the specified language."
                )
            return example_inputs
        else:
            try:
                with open(self.example_inputs_file, encoding="utf-8") as f:
                    return f.readlines()
            except Exception:
                raise ValueError("Given 'example_inputs_file' not readable.")

    @staticmethod
    def from_file(file_path: str) -> "AppConfig":
        with open(file_path) as f:
            config = AppConfig(**yaml.safe_load(f))
        return config



"utils.py"

from typing import Callable
from jinja2 import Template
from typing import Any, Mapping
from datetime import datetime
import os
from glob import glob
from github import Github, Auth, UnknownObjectException
import importlib.resources as pkg_resources
from loguru import logger

from .results import EvidenceSeekerResult
from .confirmation_aggregation.base import (
    confirmation_level
)
from .datamodels import Document

_PACKAGE_DATA_MODULE = "evidence_seeker.package_data"
_DEFAULT_MD_TEMPLATE = "templates/default_markdown.tmpl"

# TODO (refactor!): Using this function is very ideosynctratic
# since it hinges on specfici meta-data (which is
# specific to the EvSe Demo Dataset).


def get_grouped_sources(
        documents: list[Document] | None,
        confirmation_by_document: Mapping[str, float] | None
) -> dict[str, dict[str, Any]]:
    if documents is None or confirmation_by_document is None:
        return dict()
    docs_grouped_by_src_file = {}
    # Group documents by filenname
    for doc in documents:
        file_name = doc.metadata.get("file_name", None)
        if file_name is None:
            raise ValueError("No filename found in metadata")
        else:
            if file_name not in docs_grouped_by_src_file:
                docs_grouped_by_src_file[file_name] = {
                    "author": doc.metadata["author"],
                    "url": doc.metadata["url"],
                    "title": (
                        doc.metadata["title"]
                        .replace("{", "")
                        .replace("}", "")
                    ),
                    "texts": [],
                }
            docs_grouped_by_src_file[file_name]["texts"].append({
                "original_text": (
                    doc.metadata["original_text"]
                    .strip()
                    .replace("\n", " ")
                    .replace('"', "'")
                ),
                "conf": confirmation_by_document[doc.uid],
                "conf_level":
                    confirmation_level(
                        confirmation_by_document[doc.uid]
                    ).value,
                "full_text": (
                    doc.text
                    .strip()
                    .replace("\n", "  ")
                    .replace('"', "'")
                ),
            })
    # Sort texts by confidence score (highest first)
    for file_name in docs_grouped_by_src_file:
        docs_grouped_by_src_file[file_name]["texts"] = sorted(
            docs_grouped_by_src_file[file_name]["texts"],
            key=lambda item: item["conf"],
            reverse=True
        )
    return docs_grouped_by_src_file


def result_as_markdown(
    evse_result: EvidenceSeekerResult,
    translations: dict[str, str] = dict(),
    jinja2_md_template: str | Template | None = None,
    group_docs_by_sources: bool = False,
    show_documents: bool = True,
    **kwargs,
) -> str:
    # TODO: see task from `get_grouped_sources`
    # also: `claims` is, depending from `group_docs_by_sources`
    # differently structured! (refactor)
    if group_docs_by_sources:
        claims = [
            (
                claim,
                get_grouped_sources(
                    claim.documents,
                    claim.confirmation_by_document
                )
            )
            for claim in evse_result.claims
        ]
    else:
        claims = evse_result.claims
    # use simple template from package if
    # none is given
    if jinja2_md_template is None:
        template_path = pkg_resources.files(
            _PACKAGE_DATA_MODULE
        ).joinpath(_DEFAULT_MD_TEMPLATE)

        if not os.path.exists(str(template_path)):
            raise ValueError(
                "Template file not found or unreadable in package data module."
            )
        with open(str(template_path), encoding="utf-8") as f:
            jinja2_md_template = f.read()
            result_template = Template(jinja2_md_template)
    elif isinstance(jinja2_md_template, str):
        result_template = Template(jinja2_md_template)
    elif isinstance(jinja2_md_template, Template):
        result_template = jinja2_md_template
    else:
        raise ValueError(
            "The template must be of type 'template', 'str' or 'None'."
        )

    md = result_template.render(
        evse_result=evse_result,
        feedback=evse_result.feedback["binary"],
        statement=evse_result.request,
        time=evse_result.time,
        claims=claims,
        translation=translations,
        show_documents=show_documents,
        **kwargs,
    )
    return md


# TODO: Use enum type
def _current_subdir(subdirectory_construction: str | None) -> str:
    if subdirectory_construction is None:
        return ""
    now = datetime.now()
    if subdirectory_construction == "monthly":
        subdirectory_path = now.strftime("y%Y_m%m")
    elif subdirectory_construction == "weekly":
        year, week, _ = now.isocalendar()
        subdirectory_path = f"y{year}_w{week}"
    elif subdirectory_construction == "yearly":
        subdirectory_path = now.strftime("y%Y")
    elif subdirectory_construction == "daily":
        subdirectory_path = now.strftime("%Y_%m_%d")
    else:
        subdirectory_path = ""
    return subdirectory_path


def log_result(
    evse_result: EvidenceSeekerResult,
    result_dir: str = "",
    local_base: str = ".",
    subdirectory_construction: str | None = None,
    write_on_github: bool = False,
    github_token_name: str | None = None,
    repo_name: str | None = None,
    additional_markdown_log: bool = False,
    filename_without_suffix: Callable[[EvidenceSeekerResult], str] | str | None = None,
):
    # Do not log results if pipeline failed somehow
    # TODO: Better to use state field (in 'EvSeResult') by catching
    # errors and/or accessing the error codes from request
    # (refactor workflows or pipeline for this)
    if len(evse_result.claims) == 0:
        return
    if evse_result.time is None:
        raise ValueError("Request time not set in result.")
    # constructing file name
    if filename_without_suffix is None:
        ts = datetime.strptime(
            evse_result.time, "%Y-%m-%d %H:%M:%S UTC"
        ).strftime("%Y_%m_%d")
        filename_without_suffix = f"{ts}_{evse_result.uid}"
    elif callable(filename_without_suffix):
        filename_without_suffix = filename_without_suffix(evse_result)

    fn = f"{filename_without_suffix}.yaml"
    md_fn = f"{filename_without_suffix}.md"

    subdir = _current_subdir(subdirectory_construction)
    if write_on_github and repo_name:
        filepath = os.path.join(result_dir, subdir, fn)
        md_filepath = os.path.join(result_dir, subdir, md_fn)
        if (
            github_token_name is None
            or github_token_name not in os.environ.keys()
        ):
            raise ValueError(
                "Github token name not set or token not"
                "found as env variable by the specified name."
            )
        auth = Auth.Token(os.environ[github_token_name])
        g = Github(auth=auth)
        repo = g.get_repo(repo_name)
        logger.info(
            "Log evidence seeker result to "
            f"{filepath} "
            f"(with additional md log)" if additional_markdown_log else ""
            f"in repo '{repo}' on github."
        )
        content = evse_result.yaml_dump(stream=None)
        md_content = result_as_markdown(evse_result)
        try:
            c = repo.get_contents(filepath)
            repo.update_file(
                filepath,
                f"Update result ({evse_result.uid})",
                content,
                c.sha
            )
            if additional_markdown_log:
                c = repo.get_contents(md_filepath)
                repo.update_file(
                    md_filepath,
                    f"Update result ({evse_result.uid})",
                    md_content,
                    c.sha
                )
        except UnknownObjectException:
            repo.create_file(
                filepath,
                f"Upload new result ({evse_result.uid})",
                content
            )
            if additional_markdown_log:
                repo.create_file(
                    md_filepath,
                    f"Upload new result ({evse_result.uid})",
                    md_content
                )
        return
    else:
        files = glob(os.path.join(local_base, result_dir, "**", fn),
                     recursive=True)
        if len(files) < 2:
            if len(files) == 0:
                filepath = os.path.join(local_base, result_dir, subdir, fn)
                md_filepath = os.path.join(
                    local_base, result_dir, subdir, md_fn
                )
                os.makedirs(os.path.join(local_base, result_dir, subdir), 
                            exist_ok=True)
            else:
                filepath = files[0]
                md_filepath = filepath.replace(".yaml", ".md")
            logger.info(
                "Log evidence seeker result to "
                f"{filepath} "
                f"(with additional md log)" if additional_markdown_log else ""
            )
            with open(filepath, encoding="utf-8", mode="w") as f:
                evse_result.yaml_dump(f)
            if additional_markdown_log:
                with open(md_filepath, encoding="utf-8", mode="w") as f:
                    f.write(result_as_markdown(evse_result))
        else:
            raise Exception("The uid of the result is not unique.")




"models.py"

from typing import Mapping
import enum
import pydantic


class Language(enum.Enum):
    DE = "German"
    EN = "English"


class StatementType(enum.Enum):
    DESCRIPTIVE = "descriptive"
    ASCRIPTIVE = "ascriptive"
    NORMATIVE = "normative"


class ConfirmationLevel(enum.Enum):
    STRONGLY_CONFIRMED = "strongly_confirmed"
    CONFIRMED = "confirmed"
    WEAKLY_CONFIRMED = "weakly_confirmed"
    INCONCLUSIVE_CONFIRMATION = "inconclusive_confirmation"
    WEAKLY_DISCONFIRMED = "weakly_disconfirmed"
    DISCONFIRMED = "disconfirmed"
    STRONGLY_DISCONFIRMED = "strongly_disconfirmed"


class Document(pydantic.BaseModel):
    text: str
    uid: str
    metadata: dict = {}


class CheckedClaim(pydantic.BaseModel):
    text: str
    negation: str | None = None
    uid: str
    statement_type: StatementType | None = None
    n_evidence: int | None = None
    average_confirmation: float | None = None
    evidential_uncertainty: float | None = None
    confirmation_level: ConfirmationLevel | None = None
    verbalized_confirmation: str | None = None
    documents: list[Document] | None = None
    confirmation_by_document: Mapping[str, float] | None = None
    metadata: dict = {}



"confirmation_analysis.py"

from llama_index.core.workflow import (
    Context,
    Event,
    StartEvent,
    StopEvent,
    Workflow,
    step,
)
from loguru import logger

import random
from typing import List, Set, Dict, Optional, Any
from llama_index.core.llms import ChatResponse
import numpy as np
import re
import enum
import json

from evidence_seeker.backend import (
    get_openai_llm,
    OpenAILikeWithGuidance
)

from .config import (
    ConfirmationAnalyzerConfig,
    LogProbsType,
    GuidanceType,
    ConfirmationAnalyzerModelStepConfig
)


class BranchType(enum.Enum):
    CLAIM_BRANCH = "claim_branch"
    NEGATION_BRANCH = "negation_branch"


class FreetextConfirmationAnalysisEvent(Event):
    name: str = "freetext_confirmation_analysis"
    statement: str
    evidence_item: str
    branch: BranchType


class MultipleChoiceConfirmationAnalysisEvent(Event):
    name: str = "multiple_choice_confirmation_analysis"
    statement: str
    evidence_item: str
    freetext_confirmation_analysis: str
    branch: BranchType


class CollectAnalysesEvent(Event):
    """Marks aggregation of branched analyses."""
    prob_claim_entailed: Optional[float]
    branch: BranchType


class SimpleConfirmationAnalysisWorkflow(Workflow):
    "Simple confirmation analysis workflow."
    def __init__(self, config: ConfirmationAnalyzerConfig, **kwargs):
        if "timeout" not in kwargs:
            kwargs["timeout"] = config.timeout
        if "verbose" not in kwargs:
            kwargs["verbose"] = config.verbose
        super().__init__(**kwargs)
        self.config = config
        self.models = dict()

        mcq_config = self.config.get_step_config(
            "multiple_choice_confirmation_analysis"
        )
        self.n_repetitions_mcq = mcq_config.n_repetitions_mcq
        if (
            mcq_config.n_repetitions_mcq < 100
            and mcq_config.logprobs_type == LogProbsType.ESTIMATE.value
        ):
            logger.warning(
                "For reliably estimating log probs (LogProbsType.ESTIMATE) "
                "you should set `n_repetitions_mcq >= 100`!"
            )
        logger.debug(
            f"Using {self.n_repetitions_mcq} repetitions for "
            "multiple choice confirmation entailement question."
        )
        logger.debug(
            f"Using {mcq_config.guidance_type} as "
            "guidance type for multiple choice entailement question."
        )
        logger.debug(
            f"Using {mcq_config.logprobs_type} as "
            "logits retrieval type."
        )


    def _get_model(
        self,
        model_key: str
    ) -> OpenAILikeWithGuidance:
        if self.models.get(model_key) is None:
            model_kwargs = self.config.models[model_key]
            self.models[model_key] = get_openai_llm(**model_kwargs)
        return self.models[model_key]

    @step
    async def start(
        self, ctx: Context, ev: StartEvent
    ) -> FreetextConfirmationAnalysisEvent:
        # Analysis for the claim
        ctx.send_event(
            FreetextConfirmationAnalysisEvent(
                statement=ev.clarified_claim.text,
                evidence_item=ev.evidence_item,
                branch=BranchType.CLAIM_BRANCH,
            )
        )
        # Analysis for the claim's negation
        ctx.send_event(
            FreetextConfirmationAnalysisEvent(
                statement=ev.clarified_claim.negation,
                evidence_item=ev.evidence_item,
                branch=BranchType.NEGATION_BRANCH,
            )
        )

    @step
    async def freetext_analysis(
        self, ctx: Context, ev: FreetextConfirmationAnalysisEvent
    ) -> MultipleChoiceConfirmationAnalysisEvent:

        chat_template = self.config.get_chat_template(ev.name)
        messages = chat_template.format_messages(
            statement=ev.statement, evidence_item=ev.evidence_item
        )
        # TODO: check if it works with step-specific models
        model_key = self.config.get_model_key(ev.name)
        llm = self._get_model(model_key)
        response = await llm.achat(messages=messages)

        for i in range(self.n_repetitions_mcq):
            ctx.send_event(
                MultipleChoiceConfirmationAnalysisEvent(
                    statement=ev.statement,
                    evidence_item=ev.evidence_item,
                    freetext_confirmation_analysis=response.message.content,
                    # repassing branch tag to indicate workflow branch
                    branch=ev.branch,
                )
            )

    @step
    async def multiple_choice(
        self, ctx: Context, ev: MultipleChoiceConfirmationAnalysisEvent
    ) -> CollectAnalysesEvent:
        model_specific_conf = self.config.get_step_config(ev.name)

        # Randomize the answer options
        randomized_answer_options = RandomlyOrderedAnswerOptions(
            answer_options=set(model_specific_conf.answer_options),
            answer_labels=model_specific_conf.answer_labels,
            delim_str=model_specific_conf.delim_str,
        )
        # generate messages for llm
        chat_template = self.config.get_chat_template(ev.name)
        messages = chat_template.format_messages(
            statement=ev.statement,
            evidence_item=ev.evidence_item,
            freetext_confirmation_analysis=ev.freetext_confirmation_analysis,
            answer_options=randomized_answer_options.to_string(),
        )

        if model_specific_conf.logprobs_type == LogProbsType.OPENAI_LIKE.value:
            # TODO: To Check
            generation_kwargs = {
                "logprobs": True,
                "top_logprobs": 5,
            }
        else:
            generation_kwargs = dict()

        llm = self._get_model(
            self.config.get_model_key(
                "multiple_choice_confirmation_analysis"
            )
        )

        regex_str = _guidance_regex(model_specific_conf)
        response = await llm.achat_with_guidance(
            messages=messages,
            regex_str=regex_str,
            grammar_str=model_specific_conf.constrained_decoding_grammar,
            json_schema=model_specific_conf.json_schema,
            # output_cls= ...
            generation_kwargs=generation_kwargs,
            guidance_type=GuidanceType(model_specific_conf.guidance_type),
        )

        probs_dict = _get_logprobs(
            model_specific_conf.answer_labels,
            response,
            model_specific_conf,
            randomized_answer_options
        )
        if probs_dict is not None:
            prob_claim_entailed = probs_dict[model_specific_conf.claim_option]
        else:
            prob_claim_entailed = None

        return CollectAnalysesEvent(
            prob_claim_entailed=prob_claim_entailed,
            branch=ev.branch,
        )

    @step
    async def collect_analyses(
        self, ctx: Context, ev: CollectAnalysesEvent
    ) -> StopEvent:
        collected_events = ctx.collect_events(
            ev, [CollectAnalysesEvent] * 2 * self.n_repetitions_mcq
        )  # NOTE: would be nice to get rid of this magic number...
        # wait until we receive all events
        if collected_events is None:
            return None

        prob_claims = []
        prob_negation_claims = []
        for ev in collected_events:
            if ev.prob_claim_entailed is not None:
                # concatenating all results
                if ev.branch == BranchType.CLAIM_BRANCH:
                    # here, the claim_option corresponds to the claim
                    prob_claims.append(ev.prob_claim_entailed)
                elif ev.branch == BranchType.NEGATION_BRANCH:
                    # here, the claim_option corresponds to the
                    # claim's negation
                    prob_negation_claims.append(ev.prob_claim_entailed)

        if len(prob_claims) == 0 or len(prob_negation_claims) == 0:
            logger.error("Confirmation analysis failed.")
            raise ValueError("Confirmation analysis failed.")

        # calculate the confirmation score
        confirmation = np.mean(prob_claims) - np.mean(prob_negation_claims)

        return StopEvent(result=confirmation)


class RandomlyOrderedAnswerOptions():
    # Maps enumeration characters (e.g., 'A') to answer options
    enumeration_mapping: Dict[str, str]
    # a string that is used to separate the enumeration character
    # from the answer option
    delim_str: str

    def __init__(self,
                 answer_options: Set[str],
                 answer_labels: List[str] = None,
                 delim_str: Optional[str] = "."):
        """Generate a randomized list of the answer options."""
        # Shuffle the answers and enumeration characters
        shuffled_answers = list(answer_options)
        self.delim_str = delim_str
        random.shuffle(shuffled_answers)

        if answer_labels is None:
            default_enum_alphabet = list("ABCDEFGHIJKLMNOPQRSTUVWXYZ")
            answer_labels = default_enum_alphabet[:len(answer_options)]
        random.shuffle(answer_labels)

        # Create the a mapping label -> answer (e.g., 'A' -> answer option)
        self.enumeration_mapping = {
            enum: answer for enum, answer in zip(answer_labels,
                                                 shuffled_answers)
        }

    def to_string(self) -> str:
        """Generate a string representation of the answer options"""
        answer_options = ""
        delim_str = self.delim_str if self.delim_str else ''
        for label, answer in self.enumeration_mapping.items():
            answer_options += f"{label}{delim_str} {answer}\n"
        return answer_options

    def label_to_answer(
        self,
        label: str
    ) -> str:
        """Map an enumeration character to the corresponding answer option."""
        if label not in self.enumeration_mapping:
            raise ValueError(
                f"Invalid answer label: {label}\n"
                f"Must be one of {list(self.enumeration_mapping.keys())}"
            )
        return self.enumeration_mapping[label]


# ==helper functions==
def _extract_answer_label(
    answer_labels: list[str],
    chat_response: ChatResponse,
    model_specific_conf: ConfirmationAnalyzerModelStepConfig
) -> Optional[str]:
    """
    Returns the answer label based on the chat response
    of a `MultipleChoiceConfirmationAnalysisEvent`.
    Args:
        answer_labels (List): A list of the possible answer labels.
        chat_response (ChatResponse): The chat response object.
    Returns:
        Optional[str]: The answer label or `None` if it could
            not extracted.

    """
    if model_specific_conf.guidance_type == GuidanceType.PROMPTED.value:
        # validate the response
        validation_regex = model_specific_conf.validation_regex
        match = re.search(validation_regex, chat_response.message.content)
        if match:
            return match.group(1)
        else:
            msg = (
                f"The response content ({chat_response.message.content}) "
                "does not match the validation regex."
            )
            logger.warning(msg)
            return None
            # raise ValueError(msg)
    elif (
        model_specific_conf.guidance_type == GuidanceType.REGEX.value
        or model_specific_conf.guidance_type == GuidanceType.GRAMMAR.value
    ):
        return chat_response.message.content
    elif (
        model_specific_conf.guidance_type == GuidanceType.JSON.value
        or model_specific_conf.guidance_type == GuidanceType.PYDANTIC.value
    ):
        res = json.loads(chat_response.raw.choices[0].message.content)
        if "answer" not in res.keys() or res["answer"] not in set(answer_labels):
            msg = (
                f"The response content ({chat_response.raw.choices[0].message.content}) "
                "does not match the JSON format."
            )
            logger.warning(msg)
            return None 
        return res["answer"]  
    else:
        raise NotImplementedError(
            "Extracting answer label for guidance type "
            f"{model_specific_conf.guidance_type}"
            " is not implemented yet."
        )


def _get_logprobs(
        answer_labels: list[str],
        chat_response: ChatResponse,
        model_specific_conf: ConfirmationAnalyzerModelStepConfig,
        randomized_answer_options: RandomlyOrderedAnswerOptions
) -> Optional[dict[str, float]]:
    """
    Determines the log probabilities of answer options
    from the chat response of a `MultipleChoiceConfirmationAnalysisEvent`.

    Returns:
        dict: A dictionary with normalized probabilities for each option.
    """
    if model_specific_conf.logprobs_type == LogProbsType.OPENAI_LIKE.value:
        # OpenAI-like logprobs
        return _extract_logprobs(
            answer_labels,
            chat_response,
            model_specific_conf,
            randomized_answer_options
        )
    elif model_specific_conf.logprobs_type == LogProbsType.ESTIMATE.value:
        # mapping: answer (not the label) -> prob
        # initialize the mapping with prob=0 for all answers
        mapping_answer_probs = {
            answer: 0 for
            answer in randomized_answer_options.enumeration_mapping.values()
        }
        answer_label = _extract_answer_label(
            answer_labels,
            chat_response,
            model_specific_conf
        )
        # setting the prob=1 for the answer
        # that was chosen by the model
        if answer_label:
            answer = randomized_answer_options.label_to_answer(answer_label)
            mapping_answer_probs[answer] = 1.0
            return mapping_answer_probs
        else:
            logger.warning("Could not determine log probabilities.")
            return None
    else:
        raise ValueError(
            f"Unknown logprobs type: {model_specific_conf.logprobs_type}"
        )


def _extract_logprobs(
        answer_labels: list[str],
        chat_response: ChatResponse,
        model_specific_conf: ConfirmationAnalyzerModelStepConfig,
        randomized_answer_options: RandomlyOrderedAnswerOptions
) -> dict[str, Optional[float]]:
    """
    Extracts probabilites of answer options based
    from the chat response of a `MultipleChoiceConfirmationAnalysisEvent`.
    Args:
        answer_labels (List): A list of the possible answer labels.
        chat_response (ChatResponse): The chat response object
            containing raw log probabilities.
    Returns:
        dict: A dictionary with normalized probabilities for each option.
    Raises:
        ValueError: If the claim option is not in the list of options or
            if the response does not contain log probabilities.
    Warnings:
        Logs a warning if the list of alternative first tokens is not
            equal to the given response choices.
    """
    # mapping: answer (not the label) -> prob
    # initialize the mapping with prob=0 for all answers
    mapping_answer_probs : dict[str, float | None]= {
        answer: 0.0 for
        answer in randomized_answer_options.enumeration_mapping.values()
    }
    if (
        model_specific_conf.guidance_type == GuidanceType.REGEX.value
        or model_specific_conf.guidance_type == GuidanceType.GRAMMAR.value
    ):
        if not hasattr(chat_response.raw.choices[0].logprobs, "content"):
            logger.error(
                "The response does not contain log probabilities."
            )
        top_logprobs = chat_response.raw.choices[0].logprobs.content
        first_token_top_logprobs = top_logprobs[0].top_logprobs
        tokens = [token.token for token in first_token_top_logprobs]
        if set(tokens) != set(answer_labels):
            logger.warning(
                f"WARNING: The list of alternative first tokens ({tokens}) is "
                f"not equal to the given response choices ({answer_labels}). "
                "Perhaps, the constrained decoding does not work as expected."
            )
        if not set(answer_labels).issubset(set(tokens)):
            raise RuntimeError(
                f"The response choices ({answer_labels}) are not in the list "
                f"of alternative first tokens ({tokens}). "
                "Perhaps, the constrained decoding does not work as expected."
            )
        # mapping: answer label -> label probability
        probs_dict = {
            token.token: np.exp(token.logprob)
            for token in first_token_top_logprobs
            if token.token in answer_labels
        }
        # if necessary, normalize probs
        probs_sum = np.sum(list(probs_dict.values()))
        probs_dict = {token: float(prob / probs_sum)
                      for token, prob in probs_dict.items()}
        # update mapping_answer_probs
        for answer_label in probs_dict.keys():
            answer = randomized_answer_options.label_to_answer(answer_label)
            mapping_answer_probs[answer] = probs_dict[answer_label]
        logger.debug(mapping_answer_probs)
        return mapping_answer_probs
    # Assumes that answer labels are single and unique characters, but considers all tokens within the top_logprobs that contain an answer label,
    # i.e. does not assume that answer labels always occur as single tokens at the same position within the token string
    elif (model_specific_conf.guidance_type == GuidanceType.JSON.value):
        if not hasattr(chat_response.raw.choices[0].logprobs, "token_logprobs"):
            logger.error(
                "The response does not contain log probabilities."
            )
        
        answer = _extract_answer_label(
            answer_labels,
            chat_response,
            model_specific_conf
        )
        if not answer:
            raise RuntimeError(
                "No answer label extractable. Perhaps, the "
                "constrained decoding does not work as expected."
            )
        # mapping: answer label -> label probability
        probs_dict = {}
        json_prefix = '{"answer": "'
        prev = ""
        tokens = chat_response.raw.choices[0].logprobs.model_dump()["tokens"]
        top_logprobs = chat_response.raw.choices[0].logprobs.model_dump()["top_logprobs"]
        for i, token in enumerate(tokens):
            for alt in top_logprobs[i].keys():
                for label in answer_labels:
                    if (
                        (prev+alt).startswith(json_prefix+label)
                        and top_logprobs[i][alt] is not None
                        and alt.find(label) != -1
                    ):
                        if label not in probs_dict.keys(): 
                            probs_dict[label] = np.exp(top_logprobs[i][alt])
                        else:
                            probs_dict[label] += np.exp(top_logprobs[i][alt])
            prev += token

        # normalizing probs
        probs_sum = np.sum(list(probs_dict.values()))
        probs_dict = {token: float(prob / probs_sum)
                      for token, prob in probs_dict.items()}
        # update mapping_answer_probs
        for answer_label in probs_dict.keys():
            answer = randomized_answer_options.label_to_answer(answer_label)
            mapping_answer_probs[answer] = probs_dict[answer_label]

        return mapping_answer_probs
    else:
        raise NotImplementedError(
            "Extracting logprobs for guidance type "
            f"{model_specific_conf.guidance_type}"
            " is not implemented yet."
        )


def _guidance_regex(
        model_specific_conf: ConfirmationAnalyzerModelStepConfig
) -> str:
    # Construct the regex programmatically, if not given in config
    if model_specific_conf.constrained_decoding_regex:
        return model_specific_conf.constrained_decoding_regex
    else:
        return rf"^({'|'.join(model_specific_conf.answer_labels)})$"



#!/usr/bin/env python3

from os import path, makedirs
import sys
from loguru import logger
import argparse
import shutil  # For copying files
import importlib.resources as pkg_resources
from pathlib import Path
import asyncio

from evidence_seeker import (
    IndexBuilder,
    EvidenceSeekerResult,
    log_result
)
from evidence_seeker import EvidenceSeeker


_PACKAGE_DATA_MODULE = "evidence_seeker.package_data"
_DEFAULT_NAME = "my-evidence-seeker"
_BASE_DIR = "."

_CONFIG_DIR_NAME = "config"
_RETRIEVAL_CONFIG_FILE_NAME = "retrieval_config.yaml"
_PREPROCESSING_CONFIG_FILE_NAME = "preprocessing_config.yaml"
_CONFIRMATION_ANALYSIS_CONFIG_FILE_NAME = "confirmation_analysis_config.yaml"

_KNOWLEDGE_BASE_DIR_NAME = "knowledge_base"
_SCRIPTS_DIR_NAME = "scripts"
_INDEX_DIR_NAME = "embeddings"
_LOGS_DIR_NAME = "logs"

# Remove the default handler
logger.remove()
logger.add(
    sys.stdout,
    colorize=True,
    format=(
        "<white>{time:DD.MM.YYYY}</white> <level>{level} | {message}</level>"
    )
)
logger.add(f"{_LOGS_DIR_NAME}/evse_cli.log",
           rotation="10 MB",
           retention="10 days",)


def main():
    parser = argparse.ArgumentParser(description="EvidenceSeeker CLI tool.")
    parser.add_argument(
        "action",
        choices=['init', 'build-index', 'run','demo-app'],
        help="Action to perform")
    parser.add_argument(
        "-n", "--name",
        default=_DEFAULT_NAME,
        help="Name of the evidence seeker instance"
    )
    parser.add_argument(
        "-i", "--input",
        default=None,
        help="Input string for running the pipeline (via 'evse run')"
    )
    parser.add_argument(
        "-o", "--output",
        default=None,
        help="File for persisting the results of the  'run' action"
    )

    args = parser.parse_args()

    if args.action == "init":
        name = args.name.strip()
        if name == _DEFAULT_NAME:
            logger.warning(
                f"Using default name '{_DEFAULT_NAME}'. "
                "Consider specifying a custom name "
                "with -n or --name during `evse init`. "
            )
        target_directory = path.join(_BASE_DIR, name)

        logger.info(
            f"Initializing directory structure in {target_directory}..."
        )
        _init_directory_structure(target_directory)
        logger.info(
            "Copying default configs and scripts ..."
        )
        _copy_package_data(target_directory)
        logger.info("Initialization complete.")
    elif args.action == "build-index":
        _build_index()
    elif args.action == "run":
        input_string = args.input.strip()
        output_file = args.output.strip() if args.output else None
        if not input_string:
            logger.error(
                "No input string provided. Use -i or --input to specify input."
            )
            sys.exit(1)
        logger.info(f"Running pipeline with input: {input_string}")
        _run_pipeline(input_string, output_file)
    elif args.action == "demo-app":
        _run_demo_app()
    else:
        parser.print_help()


def _init_directory_structure(target_directory: str = "."):

    directories = [
        _CONFIG_DIR_NAME,
        _KNOWLEDGE_BASE_DIR_NAME,
        _KNOWLEDGE_BASE_DIR_NAME + "/data_files",
        _SCRIPTS_DIR_NAME,
        _INDEX_DIR_NAME,
        _LOGS_DIR_NAME,
    ]
    makedirs(target_directory, exist_ok=True)

    for dir_name in directories:
        dir_name = path.join(target_directory, dir_name)
        if not path.exists(dir_name):
            makedirs(dir_name)
            logger.info(f"Created directory: {dir_name}")
        else:
            logger.info(f"Directory already exists: {dir_name}")


def _copy_package_data(target_directory: str = "."):
    """
    Copies package data files to the specified base directory.
    """
    package_data_files = [
        "config/preprocessing_config.yaml",
        "config/retrieval_config.yaml",
        "config/confirmation_analysis_config.yaml",
        "config/api_keys.txt",
        "config/demo_app_config.yaml"
    ]

    for file_name in package_data_files:
        try:
            source_path = pkg_resources.files(
                _PACKAGE_DATA_MODULE
            ).joinpath(file_name)
            target_path = path.join(target_directory, file_name)
            # Copy the file
            if not path.exists(target_path):
                shutil.copy(source_path, target_path)
                logger.info(f"Copied {file_name} to {target_path}")
            else:
                logger.info(f"File already exists: {target_path}")
        except FileNotFoundError:
            logger.warning(
                f"Source file '{source_path}' not found in package data"
                f" module '{_PACKAGE_DATA_MODULE}'. Skipping.")
        except Exception as e:
            logger.error(
                f"Error copying '{source_path}' to '{target_path}': {e}"
            )


def _build_index(base_dir: str = "."):

    config_file = Path(base_dir, _CONFIG_DIR_NAME, _RETRIEVAL_CONFIG_FILE_NAME)

    if not config_file.exists():
        logger.error(
            f"Configuration file '{config_file}' does not exist. "
            "Please run 'evse init' first, and call this command from the "
            "root directory of your evidence seeker instance."
        )
        sys.exit(1)

    index_builder = IndexBuilder.from_config_file(str(config_file))
    index_builder.build_index()


def _run_pipeline(
        input_string: str,
        output_file: str = None,
        base_dir: str = "."
):

    config_dir = path.join(base_dir, _CONFIG_DIR_NAME)
    pipeline = EvidenceSeeker(
        retrieval_config_file=path.join(
            config_dir, _RETRIEVAL_CONFIG_FILE_NAME
        ),
        confirmation_analysis_config_file=path.join(
            config_dir, _CONFIRMATION_ANALYSIS_CONFIG_FILE_NAME
        ),
        preprocessing_config_file=path.join(
            config_dir, _PREPROCESSING_CONFIG_FILE_NAME
        )
    )

    results = asyncio.run(pipeline(input_string))

    if results:
        logger.info("Pipeline executed successfully.")
        # write results to a file

        evse_result = EvidenceSeekerResult(
            retrieval_config=pipeline.retriever.config,
            confirmation_config=pipeline.analyzer.config,
            preprocessing_config=pipeline.preprocessor.config,
            request=input_string,
            claims=results,
        )
        log_result(
            evse_result=evse_result,
            result_dir=_LOGS_DIR_NAME,
            local_base=base_dir,
            additional_markdown_log=True,
        )
    else:
        logger.warning(
            "No results returned from the pipeline. "
            "Please check your input and configuration."
        )


def _run_demo_app(base_dir: str = "."):
    """Launch the Gradio demo app"""
    try:
        # Set environment variables for the app
        import os
        config_file = path.join(
            base_dir,
            _CONFIG_DIR_NAME,
            "demo_app_config.yaml"
        )
        if not path.exists(config_file):
            logger.error(
                f"Demo app configuration file '{config_file}' does not exist. "
                "Please run 'evse init' first, and ensure demo_app_config.yaml exists."
            )
            sys.exit(1)
        else:
            logger.info(f"Using the following config file: {config_file}")
        # Set config path for the app
        os.environ["APP_CONFIG_FILE"] = config_file

        logger.info("Starting Gradio demo app...")
        logger.info("The app will be available at http://localhost:7860")
        # Import the demo app
        from evidence_seeker.demo_app.app import evse_demo_app

        # Launch the app
        evse_demo_app.launch(
            server_name="0.0.0.0",  # Allow external access
            server_port=7860,       # Default Gradio port
            share=False,            # Set to True for public sharing
            debug=False,            # Set to True for development
            show_error=True         # Show detailed errors
        )

    except ImportError as e:
        logger.error(f"Failed to import demo app: {e}")
        logger.error("Make sure the demo app dependencies are installed.")
        sys.exit(1)
    except Exception as e:
        logger.error(f"Failed to start demo app: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()



"results.py"

from typing import Any
import pydantic
import yaml
import numpy as np
import uuid
from datetime import datetime, timezone


from evidence_seeker.preprocessing.config import ClaimPreprocessingConfig
from evidence_seeker.retrieval.config import RetrievalConfig
from evidence_seeker.confirmation_analysis.config import ConfirmationAnalyzerConfig
from evidence_seeker.datamodels import (
    StatementType,
    ConfirmationLevel,
    CheckedClaim
)


class EvidenceSeekerResult(pydantic.BaseModel):
    uid: str = pydantic.Field(
        default_factory=lambda: str(uuid.uuid4())
    )
    request: str | None = None
    time: str = pydantic.Field(
        default_factory=lambda: datetime.now(
            timezone.utc
        ).strftime("%Y-%m-%d %H:%M:%S UTC"))
    retrieval_config: RetrievalConfig
    confirmation_config: ConfirmationAnalyzerConfig
    preprocessing_config: ClaimPreprocessingConfig
    claims: list[CheckedClaim] = []
    feedback: dict[str, Any] = {
        # TODO: perhaps better with an enum.Enum?
        "binary": None
    }

    def yaml_dump(self, stream) -> None | str | bytes:
        yaml.add_representer(
            StatementType,
            representer=(
                lambda dumper, data: dumper.represent_str(
                    data.value
                )
            )
        )
        yaml.add_representer(
            ConfirmationLevel,
            representer=(
                lambda dumper, data: dumper.represent_str(
                    data.value
                )
            )
        )
        yaml.add_representer(
            np.ndarray,
            representer=(
                lambda dumper, data: dumper.represent_sequence(
                    "!nparray", [float(x) for x in data]
                )
            ),
        )
        yaml.add_representer(
            np.float64,
            representer=(lambda dumper, data: dumper.represent_float(float(data))),
        )
        return yaml.dump(
            self.model_dump(),
            stream,
            allow_unicode=True,
            default_flow_style=False,
            encoding="utf-8",
        )

    @classmethod
    def from_logfile(cls, path) -> "EvidenceSeekerResult":
        # yaml.add_constructor(
        #     "!python/object/apply:evidence_seeker.datamodels.StatementType",
        #     constructor=(lambda _, node: StatementType(node.value[0].value)),
        # )
        # yaml.add_constructor(
        #     "tag:yaml.org,2002:python/object/apply:evidence_seeker.datamodels.StatementType",
        #     constructor=(lambda _, node: StatementType(node.value[0].value)),
        # )
        yaml.add_constructor(
            "!nparray",
            constructor=(
                lambda _, node: np.array([float(n.value) for n in node.value])
            ),
        )
        with open(path, encoding="utf-8") as f:
            res = yaml.full_load(f)
        # model_validate will handle str2enum convertion
        return cls.model_validate(res)

    def count_claims(self) -> dict[str, int]:
        return {
            x: [c.statement_type.value for c in self.claims
                if c.statement_type is not None].count(x)
            for x in ["normative", "descriptive", "ascriptive"]
        }



"preprocessing_workflow"

from typing import List

# import json
from llama_index.core.workflow import (
    Context,
    Event,
    StartEvent,
    StopEvent,
    Workflow,
    step,
)
from loguru import logger
from pydantic import BaseModel, Field
import uuid

from evidence_seeker.preprocessing.config import (
    ClaimPreprocessingConfig,
)
from evidence_seeker.datamodels import CheckedClaim, Language, StatementType
from evidence_seeker.backend import get_openai_llm, OpenAILikeWithGuidance


# ==pydantic models for constrained decoding==


class Claims(BaseModel):
    """A list of claims."""

    claims: List[str] = Field(description="A list of claims.")


# ==events==


class DescriptiveAnalysisEvent(Event):
    name: str = "freetext_descriptive_analysis"
    claim: str


class ListDescriptiveClaimsEvent(Event):
    name: str = "list_descriptive_statements"
    claim: str
    descriptive_analysis: str


class AscriptiveAnalysisEvent(Event):
    name: str = "freetext_ascriptive_analysis"
    claim: str


class ListAscriptiveClaimsEvent(Event):
    name: str = "list_ascriptive_statements"
    claim: str
    ascriptive_analysis: str


class NormativeAnalysisEvent(Event):
    name: str = "freetext_normative_analysis"
    claim: str


class ListNormativeClaimsEvent(Event):
    name: str = "list_normative_statements"
    claim: str
    normative_analysis: str


class NegateClaimEvent(Event):
    name: str = "negate_claim"
    statement: str
    statement_type: str


class StartedNegatingClaims(Event):
    """
    Marks that corresp. num_claims_to_negate has been set.
    Needed for buffering.
    """


class CollectClarifiedClaimsEvent(Event):
    clarified_claim: CheckedClaim


# ==workflow==


class PreprocessingWorkflow(Workflow):
    """
    This workflow lists claims based on seperated analyses. For instance,
    the list of descriptive statements is based on the free text analysis
    of descriptive content (and does not consider the results of the other
    free text analyses).
    """

    def __init__(self, config: ClaimPreprocessingConfig, **kwargs):
        if "timeout" not in kwargs:
            kwargs["timeout"] = config.timeout
        if "verbose" not in kwargs:
            kwargs["verbose"] = config.verbose
        super().__init__(**kwargs)
        self.config = config
        self.lang = Language._member_map_.get(config.language)
        if not self.lang:
            raise ValueError(f"Language {config.language} not supported.")
        self.models = dict()

    def _get_model(
        self,
        model_key: str
    ) -> OpenAILikeWithGuidance:
        if self.models.get(model_key) is None:
            model_kwargs = self.config.models[model_key]
            self.models[model_key] = get_openai_llm(**model_kwargs)
        return self.models[model_key]

    @step
    async def start(
        self, ctx: Context, ev: StartEvent
    ) -> (
        NormativeAnalysisEvent
        | DescriptiveAnalysisEvent
        | AscriptiveAnalysisEvent
    ):
        ctx.send_event(DescriptiveAnalysisEvent(claim=ev.claim))
        ctx.send_event(AscriptiveAnalysisEvent(claim=ev.claim))
        ctx.send_event(NormativeAnalysisEvent(claim=ev.claim))

    @step
    async def descriptive_analysis(
        self, ctx: Context, ev: DescriptiveAnalysisEvent
    ) -> ListDescriptiveClaimsEvent:

        logger.debug(f"Analysing descriptive aspects of claim '{ev.claim}'.")
        model_key = self.config.get_model_key(ev.name)
        llm = self._get_model(model_key)
        chat_template = self.config.get_chat_template(ev.name)
        messages = chat_template.format_messages(
            claim=ev.claim,
            language=self.lang.value
        )
        response = await llm.achat(messages=messages)

        return ListDescriptiveClaimsEvent(
            claim=ev.claim,
            descriptive_analysis=response.message.content,
        )

    @step
    async def list_descriptive_claims(
        self, ctx: Context, ev: ListDescriptiveClaimsEvent
    ) -> NegateClaimEvent | StartedNegatingClaims:

        model_key = self.config.get_model_key(ev.name)
        llm = self._get_model(model_key)
        # json_schema = json.dumps(Claims.model_json_schema(), indent=2)
        chat_template = self.config.get_chat_template(ev.name)
        messages = chat_template.format_messages(
            claim=ev.claim,
            descriptive_analysis=ev.descriptive_analysis,
            language=self.lang.value,
        )
        response = await llm.achat_with_guidance(
            messages=messages,
            # json_schema=json_schema,
            output_cls=Claims
        )
        claims = Claims.model_validate_json(response.message.content)
        # deduplictae claims
        claims.claims = list(set(claims.claims))

        # store number of descriptive claims to negate in the context
        await ctx.set("num_descriptive_claims", len(claims.claims))
        logger.debug(
            f"Identified {len(claims.claims)} decriptive interpretations."
        )

        for claim in claims.claims:
            ctx.send_event(
                NegateClaimEvent(
                    statement=claim,
                    statement_type=StatementType.DESCRIPTIVE.value,
                )
            )

        return StartedNegatingClaims()

    @step
    async def ascriptive_analysis(
        self, ctx: Context, ev: AscriptiveAnalysisEvent
    ) -> ListAscriptiveClaimsEvent:

        logger.debug(f"Analysing ascriptive aspects of claim '{ev.claim}'.")
        model_key = self.config.get_model_key(ev.name)
        llm = self._get_model(model_key)
        chat_template = self.config.get_chat_template(ev.name)
        messages = chat_template.format_messages(
            claim=ev.claim,
            language=self.lang.value
        )
        response = await llm.achat(messages=messages)

        return ListAscriptiveClaimsEvent(
            claim=ev.claim,
            ascriptive_analysis=response.message.content,
        )

    @step
    async def list_ascriptive_claims(
        self, ctx: Context, ev: ListAscriptiveClaimsEvent
    ) -> NegateClaimEvent | StartedNegatingClaims:

        model_key = self.config.get_model_key(ev.name)
        llm = self._get_model(model_key)
        chat_template = self.config.get_chat_template(ev.name)
        messages = chat_template.format_messages(
            claim=ev.claim,
            ascriptive_analysis=ev.ascriptive_analysis,
            language=self.lang.value,
        )
        response = await llm.achat_with_guidance(
            messages=messages,
            # json_schema=json_schema,
            output_cls=Claims
        )
        claims = Claims.model_validate_json(response.message.content)
        # deduplictae claims
        claims.claims = list(set(claims.claims))

        # store number of ascriptive claims to negate in the context
        await ctx.set("num_ascriptive_claims", len(claims.claims))
        logger.debug(
            f"Identified {len(claims.claims)} ascriptive interpretations."
        )
        for claim in claims.claims:
            ctx.send_event(
                NegateClaimEvent(
                    statement=claim,
                    statement_type=StatementType.ASCRIPTIVE.value
                )
            )

        return StartedNegatingClaims()

    @step
    async def normative_analysis(
        self, ctx: Context, ev: NormativeAnalysisEvent
    ) -> ListNormativeClaimsEvent:

        logger.debug(f"Analysing normative aspects of claim: '{ev.claim}'.")
        model_key = self.config.get_model_key(ev.name)
        llm = self._get_model(model_key)
        chat_template = self.config.get_chat_template(ev.name)
        messages = chat_template.format_messages(
            claim=ev.claim,
            language=self.lang.value
        )
        response = await llm.achat(messages=messages)

        return ListNormativeClaimsEvent(
            claim=ev.claim,
            normative_analysis=response.message.content,
        )

    @step
    async def list_normative_claims(
        self, ctx: Context, ev: ListNormativeClaimsEvent
    ) -> NegateClaimEvent | StartedNegatingClaims:

        model_key = self.config.get_model_key(ev.name)
        llm = self._get_model(model_key)
        chat_template = self.config.get_chat_template(ev.name)
        messages = chat_template.format_messages(
            claim=ev.claim,
            normative_analysis=ev.normative_analysis,
            language=self.lang.value
        )
        response = await llm.achat_with_guidance(
            messages=messages,
            # json_schema=json_schema,
            output_cls=Claims
        )
        claims = Claims.model_validate_json(response.message.content)
        # deduplictae claims
        claims.claims = list(set(claims.claims))

        # store number of normative claims to negate in the context
        await ctx.set("num_normative_claims", len(claims.claims))
        logger.debug(
            f"Identified {len(claims.claims)} normative interpretations."
        )

        for claim in claims.claims:
            ctx.send_event(
                NegateClaimEvent(
                    statement=claim,
                    statement_type=StatementType.NORMATIVE.value
                )
            )

        return StartedNegatingClaims()

    @step(num_workers=16)
    async def negate_claim(
        self, ctx: Context, ev: NegateClaimEvent
    ) -> CollectClarifiedClaimsEvent:
        model_key = self.config.get_model_key(ev.name)
        llm = self._get_model(model_key)
        chat_template = self.config.get_chat_template(ev.name)
        messages = chat_template.format_messages(
            statement=ev.statement,
            language=self.lang.value
        )
        response = await llm.achat(messages=messages)

        clarified_claim = CheckedClaim(
            text=ev.statement,
            negation=response.message.content,
            uid=str(uuid.uuid4()),
            statement_type=StatementType(ev.statement_type),
        )
        return CollectClarifiedClaimsEvent(clarified_claim=clarified_claim)

    @step
    async def collect_clarified_claims(
        self, ctx: Context,
        ev: CollectClarifiedClaimsEvent | StartedNegatingClaims
    ) -> StopEvent:
        num_descriptive_claims = await ctx.get("num_descriptive_claims", 0)
        num_ascriptive_claims = await ctx.get("num_ascriptive_claims", 0)
        num_normative_claims = await ctx.get("num_normative_claims", 0)
        claims_to_collect = (
            num_descriptive_claims
            + num_ascriptive_claims
            + num_normative_claims
        )
        results = ctx.collect_events(
            ev,
            [CollectClarifiedClaimsEvent] * claims_to_collect
            + [StartedNegatingClaims] * 3,
        )
        if results is None:
            return None

        clarified_claims = []
        for res in results:
            if isinstance(res, StartedNegatingClaims):
                continue
            clarified_claims.append(res.clarified_claim)

        return StopEvent(result=clarified_claims)




import os
from typing import Type, Optional, Any, Dict

from dotenv import load_dotenv
import enum
from llama_index.llms.openai_like import OpenAILike
from loguru import logger
import pydantic


class BackendType(enum.Enum):
    NIM = "nim"
    TGI = "tgi"
    OPENAI = "openai"


class GuidanceType(enum.Enum):
    JSON = "json"
    REGEX = "regex"
    GRAMMAR = "grammar"
    PYDANTIC = "pydantic"
    PROMPTED = "prompted"
    STRUCTURED_LLM = "structured_llm"


class OpenAILikeWithGuidance(OpenAILike):

    backend_type: Optional[str] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.backend_type: str = kwargs.get("backend_type", "openai")
        if self.backend_type not in [bt.value for bt in BackendType]:
            logger.warning(
                f"Unknown backend type {self.backend_type}."
                "Assuming default backend type 'openai' for guided generation."
            )
            self.backend_type = BackendType.OPENAI.value

    async def achat_with_guidance(
            self,
            messages: list[str],
            json_schema: str | Dict[str, Any] | None = None,
            output_cls: Type[pydantic.BaseModel] = None,
            regex_str: str | None = None,
            grammar_str: str | None = None,
            generation_kwargs: dict = dict(),
            guidance_type: GuidanceType = GuidanceType.JSON,
    ):

        if _validate_guidance_params(
                json_schema=json_schema,
                output_cls=output_cls,
                regex_str=regex_str,
                grammar_str=grammar_str,
                guidance_type=guidance_type
        ):
            if guidance_type == GuidanceType.PROMPTED:
                return await self.achat(
                    messages=messages,
                    **generation_kwargs,
                )
            # Using the llama-index interface for structured output
            # https://docs.llamaindex.ai/en/stable/understanding/extraction/
            elif (
                self.backend_type == BackendType.OPENAI.value and
                guidance_type == GuidanceType.STRUCTURED_LLM
            ):
                sllm = self.as_structured_llm(output_cls)
                return await sllm.achat(
                    messages=messages,
                    **generation_kwargs
                )
            # else, we generate backend specific kwargs for the guidance type
            else:
                guidance_kwargs = self._get_guidance_kwargs(
                    guidance_type=guidance_type,
                    json_schema=json_schema,
                    output_cls=output_cls,
                    regex_str=regex_str,
                    grammar_str=grammar_str
                )
                generation_kwargs.update(guidance_kwargs)
                return await self.achat(
                    messages=messages,
                    **generation_kwargs
                )

    def _get_guidance_kwargs(
            self,
            guidance_type: GuidanceType,
            json_schema: str | Dict[str, Any] | None = None,
            output_cls: Type[pydantic.BaseModel] = None,
            regex_str: str | None = None,
            grammar_str: str | None = None,

    ) -> dict:
        """
        Get the kwargs for the guidance type.

        Raise ValueError if the guidance type is not supported by the backend.
        """
        if json_schema is None and output_cls is not None:
            json_schema = output_cls.model_json_schema()

        # depending on the backend_type and guidance type, we generate
        # the kwargs for the guided generation.

        # For NIM, we use the `extra_body`
        # https://docs.nvidia.com/nim/large-language-models/latest/structured-generation.html
        if self.backend_type == BackendType.NIM.value:
            if guidance_type == (
                GuidanceType.JSON or guidance_type == GuidanceType.PYDANTIC
            ):
                return {"extra_body": {"nvext": {"guided_json": json_schema}}}

        # for TGI (e.g., dedicated HF endpoints) we use `response_format`
        # for constrained decoding
        # https://github.com/huggingface/text-generation-inference/pull/2046
        elif self.backend_type == BackendType.TGI.value:
            if guidance_type == (
                GuidanceType.JSON or guidance_type == GuidanceType.PYDANTIC
            ):
                return {
                    "response_format": {
                        "type": "json_object",
                        "value": json_schema
                    }
                }
            if guidance_type == GuidanceType.REGEX:
                return {
                    "response_format": {
                        "type": "regex",
                        "value": regex_str
                    }
                }
        elif self.backend_type == BackendType.OPENAI.value:
            if guidance_type == (
                GuidanceType.JSON or guidance_type == GuidanceType.PYDANTIC
            ):
                return {
                    "response_format": {
                        "type": "json_schema",
                        "schema": json_schema
                    }
                }
            if guidance_type == GuidanceType.GRAMMAR:
                return {
                    "response_format": {
                        "type": "grammar",
                        "grammar": grammar_str
                    }
                }
        # else, we raise an error
        raise ValueError(
            f"Guidance type {guidance_type} is not supported by or "
            "not implemented for the backend {self.backend_type}."
        )


# TODO: Move to config classes and use pydantic validators
def _validate_guidance_params(
            json_schema: str | Dict[str, Any] | None = None,
            output_cls: Type[pydantic.BaseModel] = None,
            regex_str: str | None = None,
            grammar_str: str | None = None,
            guidance_type: GuidanceType = GuidanceType.JSON,
) -> bool:
    """
    Validate the parameters for the guidance type.
    """
    error_msg = None
    if guidance_type == GuidanceType.JSON:
        if json_schema is None and output_cls is None:
            error_msg = (
                "You should provide a JSON schema or a Pydantic class"
                " for structured output."
            )
    elif guidance_type == GuidanceType.PYDANTIC:
        if output_cls is None:
            error_msg = (
                "You should provide a Pydantic output class for "
                "structured output."
            )
    elif guidance_type == GuidanceType.REGEX:
        if regex_str is None:
            error_msg = (
                "You should provide a regex expression for "
                "constrained decoding."
            )
    elif guidance_type == GuidanceType.GRAMMAR:
        if grammar_str is None:
            error_msg = (
                "You should provide a grammar expression for "
                "constrained decoding."
            )
    elif guidance_type == GuidanceType.STRUCTURED_LLM:
        if output_cls is None:
            error_msg = (
                "You should provide a Pydantic output class "
                "for structured output."
            )
    elif guidance_type == GuidanceType.PROMPTED:
        error_msg = None
    else:
        error_msg = f"Guidance type {guidance_type} is not supported."

    if error_msg:
        raise ValueError(error_msg)
    return True


def get_openai_llm(
        api_key: str = None,
        api_key_name: str = None,
        model: str = None,
        base_url: str = None,
        backend_type: str = None,
        is_chat_model: bool = True,
        is_local: bool = False,
        is_function_calling_model: bool = False,
        context_window: int = 3900,
        max_tokens: int = 1024,
        **kwargs) -> OpenAILikeWithGuidance:

    if api_key is None and api_key_name is None:
        logger.warning(
            "Neither an API key nor a name of "
            "an env variable that holds the API key are provided. "
            "We assume the an API key is not needed an set it to "
            " 'api_key_not_needed'."
        )
        api_key = "api_key_not_needed"
    elif api_key is None and api_key_name is not None:
        logger.debug(f"Fetching api key via env var: {api_key_name}")
        api_key = os.getenv(api_key_name, None)
        if api_key is None:
            raise ValueError(
                f"The api key name {api_key_name} is not set as env variable."
            )
    elif api_key is not None and api_key_name is not None:
        logger.warning(
            "Both an API key and a name of an env variable that holds the API key "
            "are provided. We use the API key and ignore the env variable."
        )


    logger.debug(
        f"Instantiating OpenAILike model (model: {model},"
        f"base_url: {base_url})."
    )
    llm = OpenAILikeWithGuidance(
        model=model,
        api_base=base_url,
        backend_type=backend_type,
        api_key=api_key,
        is_chat_model=is_chat_model,
        is_local=is_local,
        is_function_calling_model=is_function_calling_model,
        context_window=context_window,
        max_tokens=max_tokens,
        **kwargs
    )
    return llm



"retrieval config"

import pydantic
from pydantic import model_validator
from loguru import logger
import enum

from evidence_seeker.datamodels import StatementType


class EmbedBackendType(enum.Enum):
    # Embedding via TEI (e.g., as provided by HuggingFace as a service)
    TEI = "tei"
    # Local embedding via ollama
    # TODO/TOFIX: Ollama embedding throws errors. Check if we can fix it.
    OLLAMA = "ollama"
    # Local embedding via huggingface
    HUGGINGFACE = "huggingface"
    # HF Inference API
    HUGGINGFACE_INFERENCE_API = "huggingface_inference_api"


class RetrievalConfig(pydantic.BaseModel):
    # TODO: Add field Descriptions
    config_version: str = "v0.1"
    description: str = "Configuration of EvidenceSeeker's retriever component."
    embed_base_url: str | None = None
    # https://huggingface.co/sentence-transformers/paraphrase-multilingual-mpnet-base-v2
    embed_model_name: str = "sentence-transformers/paraphrase-multilingual-mpnet-base-v2"
    embed_backend_type: str = "huggingface"
    # Used if Huggingface Inference Provider is used and billing should be
    # done via organization on Hugging Face
    # See: https://huggingface.co/docs/inference-providers/pricing (30.06.2025)
    bill_to: str | None = None
    api_key_name: str | None = None
    hub_key_name: str | None = None
    embed_batch_size: int = 32
    document_input_dir: str | None = None
    meta_data_file: str | None = None
    env_file: str | None = None
    document_input_files: list[str] | None = None
    window_size: int = 3
    index_id: str = "default_index_id"
    index_persist_path: str | None = None
    index_hub_path: str | None = None
    top_k: int = 8
    ignore_statement_types: list[str] = [StatementType.NORMATIVE.value]

    @model_validator(mode='after')
    def check_base_url(
        cls,
        config: 'RetrievalConfig'
    ) -> 'RetrievalConfig':
        if (
            not config.embed_base_url
            and (
                config.embed_backend_type == EmbedBackendType.TEI.value
                or (
                    config.embed_backend_type
                    == EmbedBackendType.HUGGINGFACE_INFERENCE_API.value
                )
            )
        ):
            msg = (
                "'embed_base_url' must be set for the selected "
                "embed_backend_type. Please provide a valid URL."
            )
            logger.error(msg)
            raise ValueError(msg)

        return config

    @model_validator(mode='after')
    def check_api_token_name(
        cls,
        config: 'RetrievalConfig'
    ) -> 'RetrievalConfig':
        if (
            not config.api_key_name
            and (
                config.embed_backend_type == EmbedBackendType.TEI.value
                or (
                    config.embed_backend_type
                    == EmbedBackendType.HUGGINGFACE_INFERENCE_API.value
                )
            )
        ):
            msg = (
                f"Check whether you need an API token for your backend "
                f"('{config.embed_backend_type}'). If you need one, set an "
                "`api_key_name` in the retriever config and provide the "
                "api token as an environment variable with that name."
            )
            logger.warning(msg)
        return config

    @model_validator(mode='after')
    def check_hub_token_name(
        cls,
        config: 'RetrievalConfig'
    ) -> 'RetrievalConfig':
        if (
            not config.hub_key_name
            and not config.index_hub_path
        ):
            msg = (
                "Check whether you need a HF hub token for saving/loading "
                "your index to/from the Hugging Face Hub. "
                "If you need one, set an "
                "`hub_key_name` in the retriever config and provide the "
                "token as an environment variable with that name."
            )
            logger.warning(msg)
        return config

    @model_validator(mode='after')
    def check_index_path(
        cls,
        config: 'RetrievalConfig'
    ) -> 'RetrievalConfig':
        if (
            not config.index_persist_path
            and not config.index_hub_path
        ):
            err_msg = (
                "Either 'index_persist_path' or 'index_hub_path' must "
                "be provided to store/load the index."
            )
            logger.error(err_msg)
            raise ValueError(err_msg)
        return config

    @model_validator(mode='after')
    def load_env_file(
        cls,
        config: 'RetrievalConfig'
    ) -> 'RetrievalConfig':
        if config.env_file is None:
            logger.warning(
                "No environment file with API keys specified for retriever. "
                "Please set 'env_file' to a valid path if you want "
                "to load environment variables from a file."
            )
        else:
            # check if the env file exists
            from os import path
            if not path.exists(config.env_file):
                err_msg = (
                    f"Environment file '{config.env_file}' does not exist. "
                    "Please provide a valid path to the environment file. "
                    "Or set it to None if you don't need it and set the "
                    "API keys in other ways as environment variables."
                )
                logger.warning(err_msg)
            else:
                # load the env file
                from dotenv import load_dotenv
                load_dotenv(config.env_file)
            logger.info(
                f"Loaded environment variables from '{config.env_file}'"
            )

        return config



"evidence_seeker.py"


import asyncio
from typing import Callable
from loguru import logger

from evidence_seeker.confirmation_aggregation.base import ConfirmationAggregator
from evidence_seeker.confirmation_analysis.base import ConfirmationAnalyzer
from evidence_seeker.datamodels import CheckedClaim
from evidence_seeker.preprocessing.base import ClaimPreprocessor
from evidence_seeker.retrieval.base import DocumentRetriever


class EvidenceSeeker:

    # TODO: Refactor: Work with explicitly defined arguments
    def __init__(self, **kwargs):

        if "preprocessing_config" in kwargs:
            self.preprocessor = ClaimPreprocessor(config=kwargs["preprocessing_config"])
        elif "preprocessing_config_file" in kwargs:
            self.preprocessor = ClaimPreprocessor.from_config_file(
                kwargs["preprocessing_config_file"]
            )
        elif "preprocessor" in kwargs:
            self.preprocessor = kwargs["preprocessor"]
        else:
            raise ValueError("Found no arguments to initialize preprocessor.")

        document_file_metadata = kwargs.get("document_file_metadata")
        if (
            document_file_metadata is not None
            and not isinstance(document_file_metadata, Callable)
        ):
            logger.warning("kwarg 'document_file_metadata' must be a callable.")
            document_file_metadata = None
        if "retrieval_config" in kwargs:
            self.retriever = DocumentRetriever(
                config=kwargs["retrieval_config"], 
                document_file_metadata=document_file_metadata
            )
        elif "retrieval_config_file" in kwargs:
            self.retriever = DocumentRetriever.from_config_file(
                kwargs["retrieval_config_file"]
            )
        elif "retriever" in kwargs:
            self.retriever = kwargs["retriever"]
        else:
            self.retriever = DocumentRetriever(
                document_file_metadata=document_file_metadata
            )

        if "confirmation_analysis_config" in kwargs:
            self.analyzer = ConfirmationAnalyzer(
                config=kwargs["confirmation_analysis_config"]
            )
        elif "confirmation_analysis_config_file" in kwargs:
            self.analyzer = ConfirmationAnalyzer.from_config_file(
                kwargs["confirmation_analysis_config_file"]
            )
        elif "confirmation_analyzer" in kwargs:
            self.analyzer = kwargs["confirmation_analyzer"]
        else:
            self.analyzer = ConfirmationAnalyzer()

        if "confirmation_aggregation_config" in kwargs:
            self.aggregator = ConfirmationAggregator(
                config=kwargs["confirmation_aggregation_config"]
            )
        elif "confirmation_aggregation_config_file" in kwargs:
            self.aggregator = ConfirmationAggregator.from_config_file(
                kwargs["confirmation_aggregation_config_file"]
            )
        else:
            self.aggregator = ConfirmationAggregator()

    async def execute_pipeline(self, claim: str) -> list[CheckedClaim]:
        preprocessed_claims = await self.preprocessor(claim)

        async def _chain(pclaim: CheckedClaim) -> CheckedClaim:
            for acallable in [self.retriever, self.analyzer, self.aggregator]:
                pclaim = await acallable(pclaim)
            return pclaim

        return await asyncio.gather(*[_chain(pclaim)
                                      for pclaim in preprocessed_claims])

    async def __call__(self, claim: str) -> list[CheckedClaim]:
        checked_claims = [
            claim for claim in await self.execute_pipeline(claim)
        ]
        return checked_claims



# TODO: Refactor this file to only import the public API of the package.

# from .preprocessing import (
#     ClaimPreprocessor,
#     DictInitializedEvent,
#     DictInitializedPromptEvent,
#     PreprocessingSeparateListingsWorkflow,
#     SimplePreprocessingWorkflow,
# )
# 
# from .confirmation_analysis import (
#     ConfirmationAnalyzer,
#     SimpleConfirmationAnalysisWorkflow
# )
# 
# from .backend import (
#     get_openai_llm,
#     log_msg,
# )

from .retrieval.base import (
    DocumentRetriever,
    IndexBuilder,
)

from .retrieval.config import (
    RetrievalConfig,
)

from .preprocessing.base import ClaimPreprocessor
from .preprocessing.config import (
    ClaimPreprocessingConfig,
    PreprocessorStepConfig,
    PreprocessorModelStepConfig
)
from .confirmation_analysis.base import ConfirmationAnalyzer
from .confirmation_analysis.config import (
    ConfirmationAnalyzerConfig,
    ConfirmationAnalyzerModelStepConfig,
    ConfirmationAnalyzerStepConfig,
    MultipleChoiceTaskStepConfig,
)
from .confirmation_aggregation.base import ConfirmationAggregator

from .utils import (
    result_as_markdown,
    log_result,
)


from .evidence_seeker import (
    EvidenceSeeker
)

from .datamodels import (
    CheckedClaim,
    Document,
    StatementType,
    ConfirmationLevel,
)

from .results import (
    EvidenceSeekerResult
)

# Use __all__ to let type checkers know what is part of the public API.
__all__ = [
    "EvidenceSeeker",
    "DocumentRetriever",
    "IndexBuilder",
    "RetrievalConfig",
    "ClaimPreprocessingConfig",
    "PreprocessorStepConfig",
    "PreprocessorModelStepConfig",
    "ClaimPreprocessor",
    "ConfirmationAnalyzer",
    "ConfirmationAnalyzerConfig",
    "ConfirmationAnalyzerModelStepConfig",
    "ConfirmationAnalyzerStepConfig",
    "MultipleChoiceTaskStepConfig",
    "ConfirmationAggregator",
#    "ClaimPreprocessor",
#    "DictInitializedEvent",
#    "DictInitializedPromptEvent",
#    "PreprocessingSeparateListingsWorkflow",
#    "get_openai_llm",
#    "log_msg",
#    "SimplePreprocessingWorkflow",
#    "ConfirmationAnalyzer",
    "CheckedClaim",
    "Document",
    "StatementType",
    "ConfirmationLevel",
    "EvidenceSeekerResult",
    "result_as_markdown",
    "log_result",
#    "SimpleConfirmationAnalysisWorkflow"
]



from __future__ import annotations

import gradio as gr
import random
import os
from loguru import logger
from datetime import datetime, timezone
from argon2 import PasswordHasher
from argon2.exceptions import VerificationError

from evidence_seeker.demo_app.app_config import AppConfig

from evidence_seeker import (
    EvidenceSeeker,
    EvidenceSeekerResult,
    result_as_markdown,
    log_result,
    ConfirmationLevel
)

from evidence_seeker import (
    CheckedClaim,
    Document,
    StatementType
)

# ### APP CONFIGURATION ####
# dotenv.load_dotenv()

config_file_path = os.getenv("APP_CONFIG_FILE", None)

if config_file_path is None:
    raise ValueError("Missing configruation file.")
APP_CONFIG = AppConfig.from_file(config_file_path)
ui = APP_CONFIG.ui_texts[APP_CONFIG.language]

UI_TEST_MODE = os.getenv("UI_TEST_MODE", False)

# used for UI_TEST_MODE
# TODO: provide the following dummy instance via util.py:
# dummy claims, docs and dumy evse result
_dummy_docs = [
    Document(
        text='While there is high confidence that oxygen levels have ...',
        uid='1f47ce98-4105-4ddc-98a9-c4956dab2000',
        metadata={
            'page_label': '74',
            'file_name': 'IPCC_AR6_WGI_TS.pdf',
            'author': 'IPCC Working Group I',
            'original_text': 'While there is low confidence in 20th century ...',
            'url': 'www.dummy_url.com',
            'title': 'Dummy Title'
        }
    ),
    Document(
        text='Based on recent refined \nanalyses of the ... ',
        uid='6fcd6c0f-99a1-48e7-881f-f79758c54769',
        metadata={
            'page_label': '74',
            'file_name': 'IPCC_AR6_WGI_TS.pdf',
            'author': 'IPCC Working Group I',
            'original_text': 'The AMOC was relatively stable during the past ...',
            'url': 'www.dummy_url.com',
            'title': 'Dummy Title'
        }
    ),
]

_dummy_claims = [
    CheckedClaim(
        text="The AMOC is slowing down",
        negation="The AMOC is not changing",
        uid="123",
        documents=_dummy_docs,
        n_evidence=2,
        statement_type=StatementType.DESCRIPTIVE,
        average_confirmation=0.2,
        confirmation_level=ConfirmationLevel.WEAKLY_CONFIRMED,
        evidential_uncertainty=0.1,
        verbalized_confirmation="The claim is weakly confirmed.",
        confirmation_by_document={
            "1f47ce98-4105-4ddc-98a9-c4956dab2000": 0.1,
            "6fcd6c0f-99a1-48e7-881f-f79758c54769": 0.3,
        },
    ),
]


def check_password(input_password: str, hash: str) -> bool:
    ph = PasswordHasher()
    try:
        ph.verify(hash, input_password)
        return True
    except VerificationError:
        return False


def auth(pw: str, password_authenticated: bool):
    output = ""
    b = gr.Textbox(value="")
    if APP_CONFIG.password_env_name not in os.environ:
        output = ui['server_error']
    elif not check_password(pw, os.environ[APP_CONFIG.password_env_name]):
        output = ui['wrong_password']
    else:
        output = ui['continue']
        password_authenticated = True
    return output, password_authenticated, b


def reactivate(check_btn, statement):
    if statement.strip() != "":
        check_btn = gr.Button(visible=True, interactive=True)
    good = gr.Button(visible=True, interactive=True)
    bad = gr.Button(visible=True, interactive=True)
    feedback = gr.Markdown(visible=True)
    return feedback, check_btn, good, bad


def deactivate():
    check_btn = gr.Button(interactive=False)
    good = gr.Button(interactive=False, variant="secondary", visible=False)
    bad = gr.Button(interactive=False, variant="secondary", visible=False)
    feedback = gr.Markdown(visible=False)
    return check_btn, good, bad, feedback


def log_feedback(clicked_button, last_result: EvidenceSeekerResult):
    choice = "positive" if clicked_button == "👍" else "negative"
    last_result.feedback["binary"] = (
        None if last_result.feedback["binary"] == choice else choice
    )
    logger.log("INFO", f"{last_result.feedback['binary']} feedback on results")
    return gr.Button(
        variant="primary" if last_result.feedback["binary"] else "secondary"
    ), gr.Button(variant="secondary")


def draw_example(examples: list[str]) -> str:
    random.shuffle(examples)
    return examples[0]


async def check(statement: str, last_result: EvidenceSeekerResult):
    request_time = datetime.now(timezone.utc)
    last_result.time = request_time.strftime("%Y-%m-%d %H:%M:%S UTC")
    last_result.request = statement

    if UI_TEST_MODE:
        last_result.claims = _dummy_claims
    else:
        logger.log("INFO", f"Checking '{statement}'... This could take a while.")
        checked_claims = await EVIDENCE_SEEKER(statement)
        last_result.claims = checked_claims

    result = result_as_markdown(
        evse_result=last_result,
        translations=APP_CONFIG.translations[APP_CONFIG.language],
        jinja2_md_template=APP_CONFIG.md_template,
        group_docs_by_sources=True
    )

    logger.info(
        f"Result of statement '{statement}' checked (uid: {last_result.uid})",
    )
    return result, last_result

logger.info(
    "Using the following config files to initiate pipeline:"
    f"{APP_CONFIG.retrieval_config_file}\n"
    f"{APP_CONFIG.confirmation_analysis_config_file}\n"
    f"{APP_CONFIG.preprocessing_config_file}\n"
)
EVIDENCE_SEEKER = EvidenceSeeker(
    preprocessing_config_file=APP_CONFIG.preprocessing_config_file,
    retrieval_config_file=APP_CONFIG.retrieval_config_file,
    confirmation_analysis_config_file=APP_CONFIG.confirmation_analysis_config_file,
)

with gr.Blocks(title="EvidenceSeeker") as evse_demo_app:
    last_result = gr.State(
        EvidenceSeekerResult(
            retrieval_config=EVIDENCE_SEEKER.retriever.config,
            confirmation_config=EVIDENCE_SEEKER.analyzer.config,
            preprocessing_config=EVIDENCE_SEEKER.preprocessor.config,
        )
    )
    examples = gr.State(APP_CONFIG.examples)
    password_authenticated = gr.State(
        value=False if APP_CONFIG.password_protection else True
    )
    if APP_CONFIG.force_agreement:
        allow_result_persistance = gr.State(value=False)
        read_warning = gr.State(value=False)
    else:
        allow_result_persistance = gr.State(value=True)
        read_warning = gr.State(value=True)

    good = gr.Button(value="👍", visible=False, interactive=False, render=False)
    bad = gr.Button(value="👎", visible=False, interactive=False, render=False)
    feedback = gr.Markdown(
        value=ui['feedback_question'], render=False, visible=False
    )

    @gr.render(inputs=[password_authenticated, read_warning, allow_result_persistance])
    def renderApp(
        password_authenticated_val: bool,
        read_warning_val: bool,
        allow_result_persistance_val: bool
    ):
        if password_authenticated_val and read_warning_val:
            gr.Markdown(
                f"# {ui['title']}\n"
                # f"{ui['info']}\n\n"
                # f"{ui['description']}"
            )
            gr.HTML(ui['info'])
            gr.Markdown(ui['description'])
            with gr.Row():
                statement = gr.Textbox(
                    value="",
                    label=ui['statement_label'],
                    interactive=True,
                    scale=10,
                    lines=3,
                )
                with gr.Column(scale=1):
                    example_btn = gr.Button(ui['random_example'])
                    check_btn = gr.Button(
                        ui['check_statement'],
                        interactive=False,
                    )
            result = gr.Markdown(
                "",
                min_height=80,
                show_copy_button=True
            )
            with gr.Column():
                feedback.render()
                with gr.Row():
                    good.render()
                    bad.render()

            def logging(evse_result):
                if allow_result_persistance_val:
                    log_result(
                        evse_result=evse_result,
                        result_dir=APP_CONFIG.result_dir,
                        local_base=APP_CONFIG.local_base if APP_CONFIG.local_base else "",
                        subdirectory_construction=APP_CONFIG.subdirectory_construction,
                        write_on_github=APP_CONFIG.write_on_github,
                        github_token_name=APP_CONFIG.github_token_name,
                        repo_name=APP_CONFIG.repo_name,
                        additional_markdown_log=APP_CONFIG.save_markdown,
                    )

            check_btn.click(deactivate, None, [check_btn, good, bad, feedback]).then(
                (
                    lambda: ui['checking_message']
                ),
                None,
                result,
            ).then(check, [statement, last_result], [result, last_result]).then(
                logging, last_result, None
            ).then(
                reactivate, [check_btn, statement], [feedback, check_btn, good, bad]
            )
            good.click(log_feedback, [good, last_result], [good, bad]).then(
                logging, [last_result], None
            )
            bad.click(log_feedback, [bad, last_result], [bad, good]).then(
                logging, [last_result], None
            )
            example_btn.click(fn=draw_example, inputs=examples, outputs=statement)
            statement.change(
                lambda s: (
                    gr.Button(ui['check_statement'], interactive=False)
                    if s.strip() == ""
                    else gr.Button(ui['check_statement'], interactive=True)
                ),
                statement,
                [check_btn],
            )
        elif password_authenticated_val:
            gr.Markdown(f"# {ui['privacy_title']}")
            gr.HTML(ui['disclaimer_text'])
            gr.HTML(ui['data_policy_text'])
            consent_box = gr.Checkbox(
                False,
                label=ui['consent_text'],
                info=ui['consent_info'],
            )
            agree_button = gr.Button(ui['agree_button'])
            agree_button.click(
                lambda consent_save_res: (True, consent_save_res),
                inputs=consent_box,
                outputs=[read_warning, allow_result_persistance]
            )
        else:
            gr.Markdown(f"# {ui['title']}")
            box = gr.Textbox(
                label=ui['password_label'],
                autofocus=True,
                type="password",
                submit_btn=True,
            )
            res = gr.Markdown(value="")
            box.submit(
                auth,
                inputs=[box, password_authenticated],
                outputs=[res, password_authenticated, box]
            )


evse_demo_app.launch()



"retrieval.py"

import os
import pathlib
import tempfile
from typing import Callable, Dict, List, Optional
import uuid
import yaml

from llama_index.embeddings.text_embeddings_inference import (
    TextEmbeddingsInference,
)

from llama_index.core.base.embeddings.base import BaseEmbedding
from llama_index.embeddings.huggingface import HuggingFaceEmbedding
from llama_index.embeddings.ollama import OllamaEmbedding
from llama_index.core.bridge.pydantic import Field


from llama_index.core import (
    load_index_from_storage,
    StorageContext,
    VectorStoreIndex,
)
from loguru import logger
import tenacity

from evidence_seeker.datamodels import CheckedClaim, Document
from .config import RetrievalConfig, EmbedBackendType

INDEX_PATH_IN_REPO = "index"


class PatientTextEmbeddingsInference(TextEmbeddingsInference):
    @tenacity.retry(
        stop=tenacity.stop_after_attempt(10),
        wait=tenacity.wait_exponential(multiplier=1, max=30),
    )
    def _call_api(self, texts: List[str]) -> List[List[float]]:
        result = super()._call_api(texts)
        if "error" in result:
            raise ValueError(f"Error in API response: {result['error']}")
        return result

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(10),
        wait=tenacity.wait_exponential(multiplier=1, max=30),
    )
    async def _acall_api(self, texts: List[str]) -> List[List[float]]:
        result = await super()._acall_api(texts)
        if "error" in result:
            raise ValueError(f"Error in API response: {result['error']}")
        return result


class HFTextEmbeddingsInference(TextEmbeddingsInference):

    bill_to: Optional[str] = Field(
        default=None,
        description="Organization to bill for the inference API usage."
    )

    def _call_api(self, texts: List[str]) -> List[List[float]]:
        import httpx

        headers = self._headers()
        json_data = {"inputs": texts, "truncate": self.truncate_text}

        with httpx.Client() as client:
            response = client.post(
                f"{self.base_url}/pipeline/feature-extraction",
                headers=headers,
                json=json_data,
                timeout=self.timeout,
            )

        return response.json()

    async def _acall_api(self, texts: List[str]) -> List[List[float]]:
        import httpx

        headers = self._headers()
        json_data = {"inputs": texts, "truncate": self.truncate_text}

        async with httpx.AsyncClient() as client:
            response = await client.post(
                f"{self.base_url}/pipeline/feature-extraction",
                headers=headers,
                json=json_data,
                timeout=self.timeout,
            )

        return response.json()

    def _headers(self) -> Dict[str, str]:
        headers = {"Content-Type": "application/json"}
        if self.auth_token is not None:
            if callable(self.auth_token):
                headers["Authorization"] = (
                    f"Bearer {self.auth_token(self.base_url)}"
                )
            else:
                headers["Authorization"] = f"Bearer {self.auth_token}"

        if self.bill_to is not None:
            headers["X-HF-Bill-To"] = self.bill_to

        return headers


class DocumentRetriever:
    def __init__(self, config: RetrievalConfig | None = None, **kwargs):
        if config is None:
            config = RetrievalConfig()
        self.config = config

        self.embed_model_name = config.embed_model_name
        self.embed_backend_type = config.embed_backend_type
        self.embed_base_url = config.embed_base_url
        self.embed_batch_size = config.embed_batch_size

        self.api_token = kwargs.get(
            "token",
            os.getenv(config.api_key_name or "No API_KEY_NAME_"))
        self.hub_token = kwargs.get(
            "hub_token",
            os.getenv(config.hub_key_name or "No _HUB_KEY_NAME_")
        )

        self.index_id = config.index_id
        self.index_persist_path = config.index_persist_path
        if self.index_persist_path:
            os.path.abspath(config.index_persist_path)
        self.index_hub_path = config.index_hub_path
        self.similarity_top_k = config.top_k
        self.ignore_statement_types = config.ignore_statement_types or []

        self.bill_to = config.bill_to

        self.embed_model = _get_embed_model(
            EmbedBackendType(self.embed_backend_type),
            **_get_text_embeddings_inference_kwargs(
                embed_backend_type=EmbedBackendType(self.embed_backend_type),
                embed_model_name=self.embed_model_name,
                embed_base_url=self.embed_base_url,
                embed_batch_size=self.embed_batch_size,
                token=self.api_token,
                bill_to=self.bill_to,
            )
        )
        self.index = self.load_index()

    def load_index(self):
        if not self.index_persist_path and not self.index_hub_path:
            msg = (
                "At least, either index_persist_path or index_hub_path "
                "must be provided."
            )
            logger.error(msg)
            raise ValueError(msg)

        if self.index_persist_path:
            persist_dir = self.index_persist_path
            logger.info(
                "Using index persist path: "
                f"{os.path.abspath(persist_dir)}"
            )
            if (
                not os.path.exists(self.index_persist_path)
                # empty directory check
                or not os.listdir(self.index_persist_path)
            ):
                if not self.index_hub_path:
                    raise FileNotFoundError((
                        f"Index not found at {self.index_persist_path}."
                        "Please provide a valid path and/or set "
                        "`index_hub_path`."
                    ))
                else:
                    logger.info((
                        f"Downloading index from hub at {self.index_hub_path}"
                        f"and saving to {self.index_persist_path}"
                    ))
                    self.download_index_from_hub(persist_dir)

        if not self.index_persist_path:
            logger.info(
                f"Downloading index from hub at {self.index_hub_path}..."
            )
            # storing index in temp dir
            persist_dir = self.download_index_from_hub()
            logger.info(f"Index downloaded to temp dir: {persist_dir}")

        persist_dir = os.path.join(persist_dir, INDEX_PATH_IN_REPO)
        logger.info(f"Loading index from disk at {persist_dir}")
        # rebuild storage context
        storage_context = StorageContext.from_defaults(persist_dir=persist_dir)
        # load index
        index = load_index_from_storage(
            storage_context,
            index_id=self.index_id,
            embed_model=self.embed_model
        )

        # cleanup temp dir
        if not self.index_persist_path:
            import shutil

            shutil.rmtree(persist_dir)

        return index

    def download_index_from_hub(self, persist_dir: str | None = None) -> str:

        import huggingface_hub

        HfApi = huggingface_hub.HfApi(token=self.hub_token)
        if persist_dir is None:
            persist_dir = tempfile.mkdtemp()

        HfApi.snapshot_download(
            repo_id=self.index_hub_path,
            repo_type="dataset",
            local_dir=persist_dir,
            token=self.hub_token,
        )
        return persist_dir

    async def retrieve_documents(self, claim: CheckedClaim) -> list[Document]:
        """
        retrieve top_k documents that are relevant for the claim
        and/or its negation
        """

        retriever = self.index.as_retriever(
            similarity_top_k=self.similarity_top_k
        )
        matches = await retriever.aretrieve(claim.text)
        # NOTE: We're just using the claim text for now,
        # but we could also use the claim's negation.
        # This needs to be discussed.

        documents = []

        for match in matches:
            data = match.node.metadata.copy()
            window = data.pop("window")
            documents.append(
                Document(text=window, uid=str(uuid.uuid4()), metadata=data)
            )

        return documents

    async def retrieve_pair_documents(
            self, claim: CheckedClaim
    ) -> list[Document]:
        """
        retrieve top_k documents that are relevant for the claim
        and/or its negation
        """

        retriever = self.index.as_retriever(
            similarity_top_k=self.similarity_top_k / 2
        )
        matches: list = await retriever.aretrieve(claim.text)
        matches_neg = await retriever.aretrieve(claim.negation)
        # NOTE: We're just using the claim text for now,
        # but we could also use the claim's negation.
        # This needs to be discussed.
        matches_ids = [match.node.id_ for match in matches]
        for m in matches_neg:
            if m.node.id_ in matches_ids:
                continue
            matches.append(m)
            matches_ids.append(m.node.id_)
        documents = []
        logger.info([match.node.id_ for match in matches])
        for match in matches:
            data = match.node.metadata.copy()
            window = data.pop("window")
            documents.append(
                Document(text=window, uid=str(uuid.uuid4()), metadata=data)
            )

        return documents

    async def __call__(self, claim: CheckedClaim) -> CheckedClaim:
        if claim.statement_type.value in self.ignore_statement_types:
            claim.documents = []
        else:
            claim.documents = await self.retrieve_documents(claim)
        return claim

    @staticmethod
    def from_config_file(config_file: str, **kwargs):
        path = pathlib.Path(config_file)
        config = RetrievalConfig(**yaml.safe_load(path.read_text()))
        return DocumentRetriever(config=config, **kwargs)


def _get_text_embeddings_inference_kwargs(
            embed_backend_type: EmbedBackendType = EmbedBackendType.TEI,
            embed_model_name: str | None = None,
            embed_base_url: str | None = None,
            embed_batch_size: int = 32,
            token: str | None = None,
            bill_to: str | None = None
) -> dict:
    if embed_backend_type == EmbedBackendType.HUGGINGFACE:
        return {
            "model_name": embed_model_name,
            # see https://github.com/UKPLab/sentence-transformers/issues/3212
            "token": token if token else False,
            # ToDo/Check: How to add additional arguments?
            "embed_batch_size": embed_batch_size,
        }
    elif embed_backend_type == EmbedBackendType.TEI:
        return {
            "model_name": embed_model_name,
            "base_url": embed_base_url,
            "embed_batch_size": embed_batch_size,
            "auth_token": f"Bearer {token}",
        }
    elif embed_backend_type == EmbedBackendType.OLLAMA:
        return {
            "model_name": embed_model_name,
            "base_url": embed_base_url,
            # ToDo/Check: How to add additional arguments?
            "ollama_additional_kwargs": {
                "embed_batch_size": embed_batch_size
            }
        }
    elif embed_backend_type == EmbedBackendType.HUGGINGFACE_INFERENCE_API:
        return {
            "model_name": embed_model_name,
            "base_url": embed_base_url,
            "embed_batch_size": embed_batch_size,
            "auth_token": token,
            "bill_to": bill_to,
        }
    else:
        raise ValueError(
            f"Unsupported backend type for embedding: {embed_backend_type}. "
            f"Supported types are: {[e.value for e in EmbedBackendType]}."
        )


def _get_embed_model(
        embed_backend_type: EmbedBackendType,
        **text_embeddings_inference_kwargs
) -> BaseEmbedding:
    if embed_backend_type == EmbedBackendType.OLLAMA:
        return OllamaEmbedding(
            **text_embeddings_inference_kwargs
        )
    elif embed_backend_type == EmbedBackendType.HUGGINGFACE:
        return HuggingFaceEmbedding(
            **text_embeddings_inference_kwargs
        )
    elif embed_backend_type == EmbedBackendType.TEI:
        return PatientTextEmbeddingsInference(
            **text_embeddings_inference_kwargs
        )
    elif embed_backend_type == EmbedBackendType.HUGGINGFACE_INFERENCE_API:
        # extract bill_to if provided
        bill_to = text_embeddings_inference_kwargs.pop("bill_to", None)
        embed_model = HFTextEmbeddingsInference(
            **text_embeddings_inference_kwargs,
        )
        embed_model.bill_to = bill_to
        return embed_model
    else:
        raise ValueError(
            f"Unsupported backend type for embedding: {embed_backend_type}. "
            f"Supported types are: {[e.value for e in EmbedBackendType]}."
        )


class IndexBuilder:

    def __init__(self, config: RetrievalConfig | None = None):
        if config is not None:
            self.config = config
        else:
            self.config = RetrievalConfig()

        # For building the index, we add `INDEX_PATH_IN_REPO` as subdirectory.
        if self.config.index_persist_path:
            self.config.index_persist_path = os.path.join(
                os.path.abspath(self.config.index_persist_path),
                INDEX_PATH_IN_REPO
            )

        api_token = os.getenv(self.config.api_key_name or "No API_KEY_NAME_")
        if not api_token:
            logger.warning("No API token provided for embedding model.")
        # init embed model
        self.embed_model = _get_embed_model(
            EmbedBackendType(self.config.embed_backend_type),
            **_get_text_embeddings_inference_kwargs(
                embed_backend_type=EmbedBackendType(
                    self.config.embed_backend_type
                ),
                embed_model_name=self.config.embed_model_name,
                embed_base_url=self.config.embed_base_url,
                embed_batch_size=self.config.embed_batch_size,
                token=api_token,
                bill_to=self.config.bill_to,
            )
        )

    @staticmethod
    def from_config_file(config_file: str):
        path = pathlib.Path(config_file)
        config = RetrievalConfig(**yaml.safe_load(path.read_text()))
        return IndexBuilder(config=config)

    def build_index(
            self,
            metadata_func: Callable[[str], Dict] | None = None,
    ):
        conf = self.config
        if os.path.exists(conf.index_persist_path):
            logger.warning(
                f"Index persist path {conf.index_persist_path} "
                "already exists. Exiting without building index. "
                "If you want to rebuild the index, "
                "please remove the existing sub directory 'index'"
                f" in {conf.index_persist_path}."
            )
            return
        if conf.document_input_dir and conf.document_input_files:
            logger.warning(
                "Both 'document_input_dir' and 'document_input_files' "
                " provided'. Using 'document_input_files'."
            )
            conf.document_input_dir = None
        # handle document file metadata
        if conf.meta_data_file:
            logger.info(
                f"Using metadata file {conf.meta_data_file} for documents."
            )
            if metadata_func is not None:
                logger.warning(
                    "Both callable 'metadata_func' and "
                    "'meta_data_file' are provided. "
                    "Using 'metadata_func'."
                )
            else:
                # check if metadata file exists
                if not os.path.exists(conf.meta_data_file):
                    logger.warning(
                        f"Metadata file {conf.meta_data_file} does not exist. "
                    )
                    metadata_func = None
                else:
                    import json
                    with open(conf.meta_data_file, "r") as f:
                        metadata_dict = json.load(f)

                    def metadata_func(file_name: str):
                        metadata_dict.get(file_name, {})

        if conf.document_input_dir:
            logger.debug(f"Reading documents from {conf.document_input_dir}")
        if conf.document_input_files:
            logger.debug(f"Reading documents from {conf.document_input_files}")

        from llama_index.core import SimpleDirectoryReader
        from llama_index.core.node_parser import SentenceWindowNodeParser

        logger.info("Building document index...")
        documents = SimpleDirectoryReader(
            input_dir=conf.document_input_dir,
            input_files=conf.document_input_files,
            filename_as_id=True,
            file_metadata=metadata_func,
        ).load_data()

        logger.debug("Parsing nodes...")
        nodes = SentenceWindowNodeParser.from_defaults(
            window_size=conf.window_size,
            window_metadata_key="window",
            original_text_metadata_key="original_text",
        ).get_nodes_from_documents(documents)

        logger.debug("Creating VectorStoreIndex with embeddings...")
        index = VectorStoreIndex(
            nodes,
            use_async=False,
            embed_model=self.embed_model,
            show_progress=True
        )
        index.set_index_id(conf.index_id)

        if conf.index_persist_path:
            logger.debug(f"Persisting index to {conf.index_persist_path}")
            index.storage_context.persist(conf.index_persist_path)

        if conf.index_hub_path:
            folder_path = conf.index_persist_path

            if not folder_path:
                # Save index in tmp dict
                folder_path = tempfile.mkdtemp()
                index.storage_context.persist(folder_path)

            logger.debug(f"Uploading index to hub at {conf.upload_hub_path}")

            import huggingface_hub

            hub_token = os.getenv(conf.hub_key_name or "No _HUB_KEY_NAME_")
            if not hub_token:
                logger.warning(
                    "No Hugging Face hub token found as environment variable. "
                    "Index will not be uploaded to the hub.")
                return
            HfApi = huggingface_hub.HfApi(token=hub_token)

            HfApi.upload_folder(
                repo_id=conf.index_hub_path,
                folder_path=folder_path,
                path_in_repo=INDEX_PATH_IN_REPO,
                repo_type="dataset",
            )

            if not conf.index_persist_path:
                # remove tmp folder
                import shutil

                shutil.rmtree(folder_path)



"simple_preprocessing_workflow.py"


import json
from llama_index.core.workflow import (
    StartEvent,
    StopEvent,
    Context,
    step,
)
from typing import List
from pydantic import BaseModel, Field
import uuid

from evidence_seeker.datamodels import CheckedClaim
from evidence_seeker.backend import log_msg
from evidence_seeker.workflow import (
    DictInitializedEvent,
    DictInitializedPromptEvent,
    EvidenceSeekerWorkflow
)


class NormativeAnalysisEvent(DictInitializedPromptEvent):
    event_key: str = "normative_analysis_event"


class DescriptiveAnalysisEvent(DictInitializedPromptEvent):
    event_key: str = "descriptive_analysis_event"


class AscriptiveAnalysisEvent(DictInitializedPromptEvent):
    event_key: str = "ascriptive_analysis_event"


class NegateClaimEvent(DictInitializedPromptEvent):
    event_key: str = "negate_claim_event"


class CollectClarifiedClaimsEvent(DictInitializedEvent):
    """Event of collecting statement-negation pairs."""


class ListClaimsEvent(DictInitializedPromptEvent):
    event_key: str = "list_claims_event"


class SimplePreprocessingWorkflow(EvidenceSeekerWorkflow):
    """
    This workflow lists descriptive and ascriptive claims based on all
    free-text analyses (in contrast to the
    PreprocessingSeparateListingsWorkflow).
    """
    # static class variables (used for finding the right config entries)
    workflow_key: str = "simple_preprocessing"

    @step
    async def start(
        self, ctx: Context, ev: StartEvent
    ) -> (
        NormativeAnalysisEvent
        | DescriptiveAnalysisEvent
        | AscriptiveAnalysisEvent
    ):
        ctx.send_event(
            DescriptiveAnalysisEvent(
                init_data_dict=self.config["pipeline"][self.workflow_key][
                    "workflow_events"
                ],
                request_dict={"claim": ev.claim},
            )
        )
        ctx.send_event(
            NormativeAnalysisEvent(
                init_data_dict=self.config["pipeline"][self.workflow_key][
                    "workflow_events"
                ],
                request_dict={"claim": ev.claim},
            )
        )
        ctx.send_event(
            AscriptiveAnalysisEvent(
                init_data_dict=self.config["pipeline"][self.workflow_key][
                    "workflow_events"
                ],
                request_dict={"claim": ev.claim},
            )
        )

    @step
    async def descriptive_analysis(
        self, ctx: Context, ev: DescriptiveAnalysisEvent
    ) -> ListClaimsEvent:

        log_msg("Analysing descriptive aspects of claim.")
        request_dict = await self._prompt_step(ctx, ev, **ev.request_dict)
        return ListClaimsEvent(
            init_data_dict=ev.init_data_dict,
            request_dict=request_dict,
            # result=f"Descriptive Analaysis:\n {request_dict[ev.result_key]}",
        )

    @step
    async def normative_analysis(
        self, ctx: Context, ev: NormativeAnalysisEvent
    ) -> ListClaimsEvent:

        log_msg("Analysing normative aspects of claim.")
        request_dict = await self._prompt_step(ctx, ev, **ev.request_dict)
        return ListClaimsEvent(
            init_data_dict=ev.init_data_dict,
            request_dict=request_dict,
            # result=f"Normative Analaysis:\n {request_dict[ev.result_key]}",
        )

    @step
    async def ascriptive_analysis(
        self, ctx: Context, ev: AscriptiveAnalysisEvent
    ) -> ListClaimsEvent:

        log_msg("Analysing ascriptive aspects of claim.")
        request_dict = await self._prompt_step(ctx, ev, **ev.request_dict)
        return ListClaimsEvent(
            init_data_dict=ev.init_data_dict,
            request_dict=request_dict,
            # result=f"Ascriptive Analaysis:\n {request_dict[ev.result_key]}",
        )

    @step
    async def step_list_claims(
        self,
        ctx: Context,
        ev: ListClaimsEvent
    ) -> NegateClaimEvent:

        class Claim(BaseModel):
            """A claim or statement, for constrained decoding."""

            claim: str = Field(description="The claim expressed as one sentence.")

        class Claims(BaseModel):
            """A list of claims, for constrained ecoding."""

            claims: List[Claim] = Field(description="A list of claims.")

        log_msg("Collecting analyses...")
        collected_events = ctx.collect_events(ev, [ListClaimsEvent]*3)  # NOTE: Dangerous magic number here!
        # wait until we receive the analysis events
        if collected_events is None:
            return None
        # concatenating all results
        request_dict = dict()
        for ev in collected_events:
            request_dict.update(ev.request_dict)
        # json schema for constraint decoding
        json_schema = json.dumps(Claims.model_json_schema(), indent=2)
        request_dict = await self._constraint_prompt_step(
            ctx=ctx, ev=ev,
            json_schema=json_schema,
            output_cls=Claims,
            **request_dict
        )
        # convert the json string to Claims object
        claims = Claims.model_validate_json(
            request_dict[ev.result_key]
        ).claims
        log_msg(f"Number of claims: {len(claims)}")

        await ctx.set(
            "num_claims_to_negate", len(claims)
        )
        for claim in claims:
            ctx.send_event(
                NegateClaimEvent(
                    init_data_dict=self.config["pipeline"][self.workflow_key][
                        "workflow_events"
                    ],
                    request_dict=request_dict,
                    statement=claim.claim,
                )
            )

        return None

    @step(num_workers=10)
    async def negate_claim(
        self, ctx: Context, ev: NegateClaimEvent
    ) -> CollectClarifiedClaimsEvent:

        log_msg("Negating claim.")
        request_dict = await self._prompt_step(
            ctx, ev, statement=ev.statement, **ev.request_dict
        )
        # we init a backed claim and add it to the result dict
        clarified_claim = CheckedClaim(
            text=ev.statement,
            negation=request_dict[ev.result_key],
            uid=str(uuid.uuid4()),
        )
        return CollectClarifiedClaimsEvent(
            init_data_dict=ev.init_data_dict,
            request_dict=request_dict,
            result={
                "clarified_claim": clarified_claim,
            },
        )

    @step
    async def collect_clarified_claims(
        self, ctx: Context, ev: CollectClarifiedClaimsEvent
    ) -> StopEvent:
        claims_to_collect = await ctx.get("num_claims_to_negate")
        results = ctx.collect_events(
            ev, [CollectClarifiedClaimsEvent] * claims_to_collect
        )
        if results is None:
            return None

        clarified_claims = []

        for res in results:
            clarified_claims.append(res.result["clarified_claim"])

        request_dict = ev.request_dict
        request_dict['clarified_claims'] = clarified_claims

        return StopEvent(
            result=request_dict,
        )



"PreprocessingConfig"

from typing import Any, Dict, Optional

import pydantic
from loguru import logger
from llama_index.core import ChatPromptTemplate

from evidence_seeker.backend import GuidanceType


class PreprocessorModelStepConfig(pydantic.BaseModel):
    prompt_template: str
    system_prompt: str | None = None
    # Fields used for constrained decoding
    guidance_type: Optional[str] = None

    @pydantic.field_validator('guidance_type')
    @classmethod
    def validate_guidance_type(cls, v):
        allowed_values = {GuidanceType.JSON.value}
        if (v is not None) and (v not in allowed_values):
            raise ValueError(
                f'guidance_type must be one of {allowed_values}, got {v}'
            )
        return v


class PreprocessorStepConfig(pydantic.BaseModel):
    name: str
    description: str | None = None
    used_model_key: str | None = None
    llm_specific_configs: Dict[str, PreprocessorModelStepConfig] = dict()


class ClaimPreprocessingConfig(pydantic.BaseModel):
    config_version: str = "v0.1"
    description: str = "Configuration of EvidenceSeeker's preprocessing component."
    system_prompt: str = (
        "You are a helpful assistant with outstanding expertise in critical thinking and logico-semantic analysis. \n"
        "You have a background in philosophy and experience in fact checking and debate analysis.\n"
        "You read instructions carefully and follow them precisely. You give concise and clear answers."
    )
    language: str = "DE"
    timeout: int = 900
    # Whether or not the workflow/pipeline should print additional informative messages
    # during execution.
    verbose: bool = False
    env_file: str | None = None

    @pydantic.model_validator(mode='after')
    def load_env_file(
        cls,
        config: 'ClaimPreprocessingConfig'
    ) -> 'ClaimPreprocessingConfig':
        if config.env_file is None:
            logger.warning(
                "No environment file with API keys specified for preprocessor."
                " Please set 'env_file' to a valid path if you want "
                "to load environment variables from a file."
            )
        else:
            # check if the env file exists
            from os import path
            if not path.exists(config.env_file):
                err_msg = (
                    f"Environment file '{config.env_file}' does not exist. "
                    "Please provide a valid path to the environment file. "
                    "Or set it to None if you don't need it and set the "
                    "API keys in other ways as environment variables."
                )
                logger.warning(err_msg)
            else:
                # load the env file
                from dotenv import load_dotenv
                load_dotenv(config.env_file)
            logger.info(
                f"Loaded environment variables from '{config.env_file}'"
            )

        return config

    used_model_key: str
    freetext_descriptive_analysis: PreprocessorStepConfig = pydantic.Field(
        default_factory=lambda: PreprocessorStepConfig(
            name="freetext_descriptive_analysis",
            description="Instruct the assistant to carry out free-text factual/descriptive analysis.",
            llm_specific_configs={
                "default": PreprocessorModelStepConfig(
                    prompt_template=(
                        "The following {language} claim has been submitted for fact-checking.\n\n"
                        "<claim>{claim}</claim>\n\n"
                        "Before we proceed with retrieving evidence items, we carefully analyse the claim. "
                        "Your task is to contribute to this preparatory analysis, as detailed below.\n"
                        "In particular, you should thoroughly discuss whether the claim contains or implies "
                        "factual or descriptive statements, which can be verified or falsified by empirical "
                        "observation or through scientific analysis, and which may include, for example, "
                        "descriptive reports, historical facts, or scientific claims.\n"
                        "If so, try to identify them and render them in your own words.\n"
                        "In doing so, watch out for ambiguity and vagueness in the claim. Make alternative "
                        "interpretations explicit.\n"
                        "End your analysis with a short list of all identified factual or descriptive statements in {language}. "
                        "Formulate each statement in a concise manner and such that its factual nature stands "
                        "out clearly."
                    ),
                )
            }
        )
    )
    list_descriptive_statements: PreprocessorStepConfig = pydantic.Field(
        default_factory=lambda: PreprocessorStepConfig(
            name="list_descriptive_statements",
            description="Instruct the assistant to list factual claims.",
            llm_specific_configs={
                "default": PreprocessorModelStepConfig(
                    prompt_template=(
                        "We have previously analysed the descriptive content of the following {language} claim:\n"
                        "<claim>{claim}</claim>\n"
                        "The analysis yielded the following results:\n\n"
                        "<results>\n"
                        "{descriptive_analysis}\n"
                        "</results>\n\n"
                        "Your task is to list all factual or descriptive {language} statements identified "
                        "in the previous analysis. Only include clear cases, i.e. statements that are unambiguously "
                        "factual or descriptive.\n"
                        "Format your (possibly empty) list of statements as a JSON object.\n"
                        "Do not include any other text than the JSON object."
                    ),
                    guidance_type=GuidanceType.JSON.value
                )
            },
        ),
    )
    freetext_ascriptive_analysis: PreprocessorStepConfig = pydantic.Field(
        default_factory=lambda: PreprocessorStepConfig(
            name="freetext_ascriptive_analysis",
            description="Instruct the assistant to carry out free-text ascriptions analysis.",
            llm_specific_configs={
                "default": PreprocessorModelStepConfig(
                    prompt_template=(
                        "The following {language} claim has been submitted for fact-checking.\n\n"
                        "<claim>{claim}</claim>\n\n"
                        "Before we proceed with retrieving evidence items, we carefully analyse the claim. "
                        "Your task is to contribute to this preparatory analysis, as detailed below.\n"
                        "In particular, you should thoroughly discuss whether the claim makes any explicit "
                        "ascriptions, that is, whether it explicitly ascribes a statement to a person or an "
                        "organisation (e.g., as something the person has said, believes, acts on etc.) "
                        "rather than plainly asserting that statement straightaway.\n"
                        "If so, clarify which statements are ascribed to whom exactly and in which ways.\n"
                        "In doing so, watch out for ambiguity and vagueness in the claim. Make alternative "
                        "interpretations explicit.\n"
                        "Conclude your analysis with a short list of all identified ascriptions: "
                        "Formulate each statement in a concise manner, and such that it is transparent to "
                        "whom it is attributed. Render the clarified ascriptions in {language}."
                    ),
                )
            },
        )
    )
    list_ascriptive_statements: PreprocessorStepConfig = pydantic.Field(
        default_factory=lambda: PreprocessorStepConfig(
            name="list_ascriptive_statements",
            description="Instruct the assistant to list ascriptions.",
            llm_specific_configs={
                "default": PreprocessorModelStepConfig(
                    prompt_template=(
                        "The following {language} claim has been submitted for ascriptive content analysis.\n"
                        "<claim>{claim}</claim>\n"
                        "The analysis yielded the following results:\n\n"
                        "<results>\n"
                        "{ascriptive_analysis}\n"
                        "</results>\n\n"
                        "Your task is to list all ascriptions identified in this analysis. "
                        "Clearly state each ascription as a concise {language} "
                        "statement, such that it is transparent to whom it is attributed. Only include "
                        "ascriptions that are explicitly attributed to a specific person or organisation.\n"
                        "Format your (possibly empty) list of statements as a JSON object.\n"
                        "Do not include any other text than the JSON object."
                    ),
                    guidance_type=GuidanceType.JSON.value
                )
            },
        ),
    )
    freetext_normative_analysis: PreprocessorStepConfig = pydantic.Field(
        default_factory=lambda: PreprocessorStepConfig(
            name="freetext_normative_analysis",
            description="Instruct the assistant to carry out free-text normative analysis.",
            llm_specific_configs={
                "default": PreprocessorModelStepConfig(
                    prompt_template=(
                        "The following {language} claim has been submitted for fact-checking.\n\n"
                        "<claim>{claim}</claim>\n\n"
                        "Before we proceed with retrieving evidence items, we carefully analyse the claim. "
                        "Your task is to contribute to this preparatory analysis, as detailed below.\n"
                        "In particular, you should thoroughly discuss whether the claim contains or implies "
                        "normative statements, such as value judgements, recommendations, or evaluations. "
                        "If so, try to identify them and render them in your own words.\n"
                        "In doing so, watch out for ambiguity and vagueness in the claim. Make alternative "
                        "interpretations explicit. "
                        "However, avoid reading normative content into the claim without textual evidence.\n\n"
                        "End your analysis with a short list of all identified normative statements in {language}. "
                        "Formulate each statement in a concise manner and such that its normative nature "
                        "stands out clearly."
                    ),
                )
            },
        ),
    )
    list_normative_statements: PreprocessorStepConfig = pydantic.Field(
        default_factory=lambda: PreprocessorStepConfig(
            name="list_normative_statements",
            description="Instruct the assistant to list normative claims.",
            llm_specific_configs={
                "default": PreprocessorModelStepConfig(
                    prompt_template=(
                        "The following {language} claim has been submitted for normative content analysis.\n"
                        "<claim>{claim}</claim>\n"
                        "The analysis yielded the following results:\n\n"
                        "<results>\n"
                        "{normative_analysis}\n"
                        "</results>\n\n"
                        "Your task is to list all normative statements identified in this analysis "
                        "(e.g., value judgements, recommendations, or evaluations) in {language}.\n"
                        "Format your (possibly empty) list of statements as a JSON object.\n"
                        "Do not include any other text than the JSON object."
                    ),
                    guidance_type=GuidanceType.JSON.value
                )
            },
        ),
    )
    negate_claim: PreprocessorStepConfig = pydantic.Field(
        default_factory=lambda: PreprocessorStepConfig(
            name="negate_claim",
            description="Instruct the assistant to negate a claim.",
            llm_specific_configs={
                "default": PreprocessorModelStepConfig(
                    prompt_template=(
                        "Your task is to express the opposite of the following statement in plain "
                        "and unequivocal language.\n"
                        "Please generate a single {language} sentence that clearly states the negation.\n"
                        "<statement>\n"
                        "{statement}\n"
                        "</statement>\n"
                        "Provide only the negated statement in {language} without any additional comments."
                    ),
                )
            },
        ),
    )
    models: Dict[str, Dict[str, Any]] = pydantic.Field(
        default_factory=lambda: dict()
    )

    # ==helper functions==
    def _step_config(
        self,
        step_config: Optional[PreprocessorStepConfig] = None,
        step_name: Optional[str] = None
    ) -> PreprocessorStepConfig:
        """Internal convenience function."""
        if step_config is None and step_name is None:
            raise ValueError("Either pass a step config or a name of the pipeline step")
        if step_config is None:
            if step_name == "freetext_descriptive_analysis":
                return self.freetext_descriptive_analysis
            elif step_name == "list_descriptive_statements":
                return self.list_descriptive_statements
            elif step_name == "freetext_ascriptive_analysis":
                return self.freetext_ascriptive_analysis
            elif step_name == "list_ascriptive_statements":
                return self.list_ascriptive_statements
            elif step_name == "freetext_normative_analysis":
                return self.freetext_normative_analysis
            elif step_name == "list_normative_statements":
                return self.list_normative_statements
            elif step_name == "negate_claim":
                return self.negate_claim
            else:
                raise ValueError(f"Did not found step config for {step_name}")
        else:
            return step_config

    def get_step_config(
            self,
            step_name: Optional[str] = None,
            step_config: Optional[PreprocessorStepConfig] = None
    ) -> PreprocessorModelStepConfig:
        """Get the model specific step config for the given step name.

        The requested `PreprocessorModelStepConfig` is determined by either
        the provided `step_name` or the provided `step_config`. If both
        are given, the `step_config` is used.
        """
        step_config = self._step_config(step_config, step_name)
        # used model for this step
        if step_config.used_model_key:
            model_key = step_config.used_model_key
        else:
            model_key = self.used_model_key
        # do we have a model-specific config?
        if step_config.llm_specific_configs.get(model_key):
            model_specific_conf = step_config.llm_specific_configs[model_key]
        else:
            if step_config.llm_specific_configs.get("default") is None:
                msg = (
                    f"Default step config for {step_config.name} "
                    "not found in config."
                )
                logger.error(msg)
                raise ValueError(msg)
            model_specific_conf = step_config.llm_specific_configs["default"]
        return model_specific_conf

    def get_chat_template(
            self,
            step_name: Optional[str] = None,
            step_config: Optional[PreprocessorStepConfig] = None
    ) -> ChatPromptTemplate:
        step_config = self._step_config(step_config, step_name)
        model_specific_conf = self.get_step_config(step_config=step_config)
        prompt_template = model_specific_conf.prompt_template

        return ChatPromptTemplate.from_messages(
            [
                ("system", self.get_system_prompt(step_config=step_config)),
                ("user", prompt_template),
            ]
        )

    def get_system_prompt(
            self,
            step_name: Optional[str] = None,
            step_config: Optional[PreprocessorStepConfig] = None
    ) -> str:
        """Get the system prompt for a specific step of the workflow."""
        step_config = self._step_config(step_config, step_name)
        model_specific_conf = self.get_step_config(step_config=step_config)
        if model_specific_conf.system_prompt:
            return model_specific_conf.system_prompt
        else:
            return self.system_prompt

    def get_model_key(
            self,
            step_name: Optional[str] = None,
            step_config: Optional[PreprocessorStepConfig] = None
    ) -> str:
        """Get the model key for a specific step of the workflow."""
        step_config = self._step_config(step_config, step_name)
        if step_config.used_model_key:
            return step_config.used_model_key
        else:
            return self.used_model_key
